from __future__ import annotations

from collections.abc import Iterable
from copy import copy, deepcopy
from typing import TYPE_CHECKING, Any

from ..constants import InfrahubClientMode
from ..exceptions import FeatureNotSupportedError, NodeNotFoundError, ResourceNotDefinedError, SchemaNotFoundError
from ..graphql import Mutation, Query
from ..schema import GenericSchemaAPI, RelationshipCardinality, RelationshipKind
from ..utils import compare_lists, generate_short_id
from .attribute import Attribute
from .constants import (
    ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE,
    ARTIFACT_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE,
    ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE,
    PROPERTIES_OBJECT,
)
from .related_node import RelatedNode, RelatedNodeBase, RelatedNodeSync
from .relationship import RelationshipManager, RelationshipManagerBase, RelationshipManagerSync

if TYPE_CHECKING:
    from typing_extensions import Self

    from ..client import InfrahubClient, InfrahubClientSync
    from ..context import RequestContext
    from ..schema import MainSchemaTypesAPI
    from ..types import Order


class InfrahubNodeBase:
    """Base class for InfrahubNode and InfrahubNodeSync"""

    def __init__(self, schema: MainSchemaTypesAPI, branch: str, data: dict | None = None) -> None:
        """
        Args:
            schema: The schema of the node.
            branch: The branch where the node resides.
            data: Optional data to initialize the node.
        """
        self._schema = schema
        self._data = data
        self._branch = branch
        self._existing: bool = True
        self._attribute_data: dict[str, Attribute] = {}

        # Generate a unique ID only to be used inside the SDK
        # The format if this ID is purposely different from the ID used by the API
        # This is done to avoid confusion and potential conflicts between the IDs
        self._internal_id = generate_short_id()

        self.id = data.get("id", None) if isinstance(data, dict) else None
        self.display_label: str | None = data.get("display_label", None) if isinstance(data, dict) else None
        self.typename: str | None = data.get("__typename", schema.kind) if isinstance(data, dict) else schema.kind

        self._attributes = [item.name for item in self._schema.attributes]
        self._relationships = [item.name for item in self._schema.relationships]

        self._artifact_support = hasattr(schema, "inherit_from") and "CoreArtifactTarget" in schema.inherit_from
        self._artifact_definition_support = schema.kind == "CoreArtifactDefinition"

        if not self.id:
            self._existing = False

        self._init_attributes(data)
        self._init_relationships(data)

    def get_branch(self) -> str:
        return self._branch

    def get_path_value(self, path: str) -> Any:
        path_parts = path.split("__")
        return_value = None

        # Manage relationship value lookup
        if path_parts[0] in self._schema.relationship_names:
            related_node = getattr(self, path_parts[0], None)
            if not related_node:
                return None

            try:
                peer = related_node.get()
            except (NodeNotFoundError, ValueError):
                # This can happen while batch creating nodes, the lookup won't work as the store is not populated
                # If so we cannot complete the HFID computation as we cannot access the related node attribute value
                return None

            if attribute_piece := path_parts[1] if len(path_parts) > 1 else None:
                related_node_attribute = getattr(peer, attribute_piece, None)
            else:
                return peer.hfid or peer.id

            if property_piece := path_parts[2] if len(path_parts) > 2 else None:
                return_value = getattr(related_node_attribute, property_piece, None)
            else:
                return_value = related_node_attribute

        # Manage attribute value lookup
        if path_parts[0] in self._schema.attribute_names:
            attribute = getattr(self, path_parts[0], None)
            if property_piece := path_parts[1] if len(path_parts) > 1 else None:
                return_value = getattr(attribute, property_piece, None)
            else:
                return_value = attribute

        return return_value

    def get_human_friendly_id(self) -> list[str] | None:
        if not hasattr(self._schema, "human_friendly_id"):
            return None

        if not self._schema.human_friendly_id:
            return None

        # If an HFID component is missing we assume that it is invalid and not usable for this node
        hfid_components = [self.get_path_value(path=item) for item in self._schema.human_friendly_id]
        if None in hfid_components:
            return None
        return [str(hfid) for hfid in hfid_components]

    def get_human_friendly_id_as_string(self, include_kind: bool = False) -> str | None:
        hfid = self.get_human_friendly_id()
        if not hfid:
            return None
        if include_kind:
            hfid = [self.get_kind()] + hfid
        return "__".join(hfid)

    @property
    def hfid(self) -> list[str] | None:
        return self.get_human_friendly_id()

    @property
    def hfid_str(self) -> str | None:
        return self.get_human_friendly_id_as_string(include_kind=True)

    def _init_attributes(self, data: dict | None = None) -> None:
        for attr_schema in self._schema.attributes:
            attr_data = data.get(attr_schema.name, None) if isinstance(data, dict) else None
            self._attribute_data[attr_schema.name] = Attribute(
                name=attr_schema.name, schema=attr_schema, data=attr_data
            )

    def __setattr__(self, name: str, value: Any) -> None:
        """Set values for attributes that exist or revert to normal behaviour"""
        if "_attribute_data" in self.__dict__ and name in self._attribute_data:
            self._attribute_data[name].value = value
            return

        super().__setattr__(name, value)

    def _get_request_context(self, request_context: RequestContext | None = None) -> dict[str, Any] | None:
        if request_context:
            return request_context.model_dump(exclude_none=True)

        client: InfrahubClient | InfrahubClientSync | None = getattr(self, "_client", None)
        if not client or not client.request_context:
            return None

        return client.request_context.model_dump(exclude_none=True)

    def _init_relationships(self, data: dict | None = None) -> None:
        pass

    def __repr__(self) -> str:
        if self.display_label:
            return self.display_label
        if not self._existing:
            return f"{self._schema.kind} ({self.id})[NEW]"

        return f"{self._schema.kind} ({self.id}) "

    def get_kind(self) -> str:
        return self._schema.kind

    def get_all_kinds(self) -> list[str]:
        if hasattr(self._schema, "inherit_from"):
            return [self._schema.kind] + self._schema.inherit_from
        return [self._schema.kind]

    def is_ip_prefix(self) -> bool:
        builtin_ipprefix_kind = "BuiltinIPPrefix"
        return self.get_kind() == builtin_ipprefix_kind or builtin_ipprefix_kind in self._schema.inherit_from  # type: ignore[union-attr]

    def is_ip_address(self) -> bool:
        builtin_ipaddress_kind = "BuiltinIPAddress"
        return self.get_kind() == builtin_ipaddress_kind or builtin_ipaddress_kind in self._schema.inherit_from  # type: ignore[union-attr]

    def is_resource_pool(self) -> bool:
        return hasattr(self._schema, "inherit_from") and "CoreResourcePool" in self._schema.inherit_from  # type: ignore[union-attr]

    def get_raw_graphql_data(self) -> dict | None:
        return self._data

    def _generate_input_data(  # noqa: C901
        self,
        exclude_unmodified: bool = False,
        exclude_hfid: bool = False,
        request_context: RequestContext | None = None,
    ) -> dict[str, dict]:
        """Generate a dictionary that represent the input data required by a mutation.

        Returns:
            dict[str, Dict]: Representation of an input data in dict format
        """

        data = {}
        variables = {}

        for item_name in self._attributes:
            attr: Attribute = getattr(self, item_name)
            if attr._schema.read_only:
                continue
            attr_data = attr._generate_input_data()

            # NOTE, this code has been inherited when we splitted attributes and relationships
            # into 2 loops, most likely it's possible to simply it
            if attr_data and isinstance(attr_data, dict):
                if variable_values := attr_data.get("data"):
                    data[item_name] = variable_values
                else:
                    data[item_name] = attr_data
                if variable_names := attr_data.get("variables"):
                    variables.update(variable_names)

            elif attr_data and isinstance(attr_data, list):
                data[item_name] = attr_data

        for item_name in self._relationships:
            allocate_from_pool = False
            rel_schema = self._schema.get_relationship(name=item_name)
            if not rel_schema or rel_schema.read_only:
                continue

            rel: RelatedNodeBase | RelationshipManagerBase = getattr(self, item_name)

            if rel_schema.cardinality == RelationshipCardinality.ONE and rel_schema.optional and not rel.initialized:
                data[item_name] = None
                continue

            if rel is None or not rel.initialized:
                continue

            if isinstance(rel, (RelatedNode, RelatedNodeSync)) and rel.is_resource_pool:
                # If the relatiionship is a resource pool and the expected schema is different from the one of the pool, this means we expect to get
                # a resource from the pool itself
                allocate_from_pool = rel_schema.peer != rel.peer._schema.kind

            rel_data = rel._generate_input_data(allocate_from_pool=allocate_from_pool)

            if rel_data and isinstance(rel_data, dict):
                if variable_values := rel_data.get("data"):
                    data[item_name] = variable_values
                else:
                    data[item_name] = rel_data
                if variable_names := rel_data.get("variables"):
                    variables.update(variable_names)
            elif isinstance(rel_data, list):
                data[item_name] = rel_data
            elif rel_schema.cardinality == RelationshipCardinality.MANY:
                data[item_name] = []

        if exclude_unmodified:
            data, variables = self._strip_unmodified(data=data, variables=variables)

        mutation_variables = {key: type(value) for key, value in variables.items()}

        if self.id is not None:
            data["id"] = self.id
        elif self.hfid is not None and not exclude_hfid:
            data["hfid"] = self.hfid

        mutation_payload = {"data": data}
        if context_data := self._get_request_context(request_context=request_context):
            mutation_payload["context"] = context_data

        return {
            "data": mutation_payload,
            "variables": variables,
            "mutation_variables": mutation_variables,
        }

    @staticmethod
    def _strip_unmodified_dict(data: dict, original_data: dict, variables: dict, item: str) -> None:
        data_item = data.get(item)
        if item in original_data and isinstance(original_data[item], dict) and isinstance(data_item, dict):
            for item_key in original_data[item].keys():
                for property_name in PROPERTIES_OBJECT:
                    if item_key == property_name and isinstance(original_data[item][property_name], dict):
                        if original_data[item][property_name].get("id"):
                            original_data[item][property_name] = original_data[item][property_name]["id"]
                if item_key in data[item].keys():
                    if item_key == "id" and len(data[item].keys()) > 1:
                        # Related nodes typically require an ID. So the ID is only
                        # removed if it's the last key in the current context
                        continue

                    variable_key = None
                    if isinstance(data[item].get(item_key), str):
                        variable_key = data[item][item_key][1:]

                    if original_data[item].get(item_key) == data[item].get(item_key):
                        data[item].pop(item_key)
                    elif (
                        variable_key
                        and variable_key in variables
                        and original_data[item].get(item_key) == variables.get(variable_key)
                    ):
                        data[item].pop(item_key)
                        variables.pop(variable_key)

        # TODO: I do not feel _great_ about this
        # -> I don't even know who you are (but this is not great indeed) -- gmazoyer (quoting Thanos)
        original_data_item = original_data.get(item)
        original_data_item_is_none = original_data_item is None
        if isinstance(original_data_item, dict):
            if "node" in original_data_item:
                original_data_item_is_none = original_data_item["node"] is None
            elif "id" not in original_data_item:
                original_data_item_is_none = True

        if item in data and (data_item in ({}, []) or (data_item is None and original_data_item_is_none)):
            data.pop(item)

    def _strip_unmodified(self, data: dict, variables: dict) -> tuple[dict, dict]:
        original_data = self._data or {}
        for relationship in self._relationships:
            relationship_property = getattr(self, relationship)
            if not relationship_property or relationship not in data:
                continue
            if not relationship_property.initialized and (
                not isinstance(relationship_property, RelatedNodeBase) or not relationship_property.schema.optional
            ):
                data.pop(relationship)
            elif isinstance(relationship_property, RelationshipManagerBase) and not relationship_property.has_update:
                data.pop(relationship)

        for item in original_data.keys():
            if item in data.keys():
                if data[item] == original_data[item]:
                    if attr := getattr(self, item, None):  # this should never be None, just a safety default value
                        if not isinstance(attr, Attribute) or not attr.value_has_been_mutated:
                            data.pop(item)
                    continue
                if isinstance(original_data[item], dict):
                    self._strip_unmodified_dict(data=data, original_data=original_data, variables=variables, item=item)
                    if item in self._relationships and original_data[item].get("node"):
                        relationship_data_cardinality_one = copy(original_data)
                        relationship_data_cardinality_one[item] = original_data[item]["node"]
                        self._strip_unmodified_dict(
                            data=data, original_data=relationship_data_cardinality_one, variables=variables, item=item
                        )
                        # Run again to remove the "id" key if it's the last one remaining
                        self._strip_unmodified_dict(
                            data=data, original_data=relationship_data_cardinality_one, variables=variables, item=item
                        )

        return data, variables

    @staticmethod
    def _strip_alias(data: dict) -> dict[str, dict]:
        clean = {}

        under_node = False
        data_to_clean = data
        if "node" in data:
            under_node = True
            data_to_clean = data["node"]

        for key, value in data_to_clean.items():
            if "__alias__" in key:
                clean_key = key.split("__")[-1]
                clean[clean_key] = value
            else:
                clean[key] = value

        if under_node:
            complete = {k: v for k, v in data.items() if k != "node"}
            complete["node"] = clean
            return complete
        return clean

    def _validate_artifact_support(self, message: str) -> None:
        if not self._artifact_support:
            raise FeatureNotSupportedError(message)

    def _validate_artifact_definition_support(self, message: str) -> None:
        if not self._artifact_definition_support:
            raise FeatureNotSupportedError(message)

    def generate_query_data_init(
        self,
        filters: dict[str, Any] | None = None,
        offset: int | None = None,
        limit: int | None = None,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        partial_match: bool = False,
        order: Order | None = None,
    ) -> dict[str, Any | dict]:
        data: dict[str, Any] = {
            "count": None,
            "edges": {"node": {"id": None, "hfid": None, "display_label": None, "__typename": None}},
        }

        data["@filters"] = deepcopy(filters) if filters is not None else {}

        if order:
            data["@filters"]["order"] = order

        if offset is not None:
            data["@filters"]["offset"] = offset

        if limit is not None:
            data["@filters"]["limit"] = limit

        if include and exclude:
            in_both, _, _ = compare_lists(include, exclude)
            if in_both:
                raise ValueError(f"{in_both} are part of both include and exclude")

        if partial_match:
            data["@filters"]["partial_match"] = True

        return data

    def __hash__(self) -> int:
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, (InfrahubNode, InfrahubNodeSync)):
            return NotImplemented
        return self.id == other.id

    def _relationship_mutation(self, action: str, relation_to_update: str, related_nodes: list[str]) -> str:
        related_node_str = ["{ id: " + f'"{node}"' + " }" for node in related_nodes]
        return f"""
        mutation {{
            Relationship{action}(
                data: {{
                    id: "{self.id}",
                    name: "{relation_to_update}",
                    nodes: [{", ".join(related_node_str)}]
                }}
            ) {{
                ok
            }}
        }}
        """

    def _get_attribute(self, name: str) -> Attribute:
        if name in self._attribute_data:
            return self._attribute_data[name]

        raise ResourceNotDefinedError(message=f"The node doesn't have an attribute for {name}")


class InfrahubNode(InfrahubNodeBase):
    """Represents a Infrahub node in an asynchronous context."""

    def __init__(
        self,
        client: InfrahubClient,
        schema: MainSchemaTypesAPI,
        branch: str | None = None,
        data: dict | None = None,
    ) -> None:
        """
        Args:
            client: The client used to interact with the backend.
            schema: The schema of the node.
            branch: The branch where the node resides.
            data: Optional data to initialize the node.
        """
        self._client = client

        if isinstance(data, dict) and isinstance(data.get("node"), dict):
            data = data.get("node")

        self._relationship_cardinality_many_data: dict[str, RelationshipManager] = {}
        self._relationship_cardinality_one_data: dict[str, RelatedNode] = {}

        super().__init__(schema=schema, branch=branch or client.default_branch, data=data)

    @classmethod
    async def from_graphql(
        cls,
        client: InfrahubClient,
        branch: str,
        data: dict,
        schema: MainSchemaTypesAPI | None = None,
        timeout: int | None = None,
    ) -> Self:
        if not schema:
            node_kind = data.get("__typename") or data.get("node", {}).get("__typename", None)
            if not node_kind:
                raise ValueError("Unable to determine the type of the node, __typename not present in data")
            schema = await client.schema.get(kind=node_kind, branch=branch, timeout=timeout)

        return cls(client=client, schema=schema, branch=branch, data=cls._strip_alias(data))

    def _init_relationships(self, data: dict | RelatedNode | None = None) -> None:
        for rel_schema in self._schema.relationships:
            rel_data = data.get(rel_schema.name, None) if isinstance(data, dict) else None

            if rel_schema.cardinality == "one":
                if isinstance(rel_data, RelatedNode):
                    peer_id_data: dict[str, Any] = {
                        key: value
                        for key, value in (
                            ("id", rel_data.id),
                            ("hfid", rel_data.hfid),
                            ("__typename", rel_data.typename),
                            ("kind", rel_data.kind),
                            ("display_label", rel_data.display_label),
                        )
                        if value is not None
                    }
                    if peer_id_data:
                        rel_data = peer_id_data
                    else:
                        rel_data = None
                self._relationship_cardinality_one_data[rel_schema.name] = RelatedNode(
                    name=rel_schema.name, branch=self._branch, client=self._client, schema=rel_schema, data=rel_data
                )
            else:
                self._relationship_cardinality_many_data[rel_schema.name] = RelationshipManager(
                    name=rel_schema.name,
                    client=self._client,
                    node=self,
                    branch=self._branch,
                    schema=rel_schema,
                    data=rel_data,
                )

    def __getattr__(self, name: str) -> Attribute | RelationshipManager | RelatedNode:
        if "_attribute_data" in self.__dict__ and name in self._attribute_data:
            return self._attribute_data[name]
        if "_relationship_cardinality_many_data" in self.__dict__ and name in self._relationship_cardinality_many_data:
            return self._relationship_cardinality_many_data[name]
        if "_relationship_cardinality_one_data" in self.__dict__ and name in self._relationship_cardinality_one_data:
            return self._relationship_cardinality_one_data[name]

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        """Set values for relationship names that exist or revert to normal behaviour"""
        if "_relationship_cardinality_one_data" in self.__dict__ and name in self._relationship_cardinality_one_data:
            rel_schemas = [rel_schema for rel_schema in self._schema.relationships if rel_schema.name == name]
            if not rel_schemas:
                raise SchemaNotFoundError(
                    identifier=self._schema.kind,
                    message=f"Unable to find relationship schema for '{name}' on {self._schema.kind}",
                )
            rel_schema = rel_schemas[0]
            self._relationship_cardinality_one_data[name] = RelatedNode(
                name=rel_schema.name, branch=self._branch, client=self._client, schema=rel_schema, data=value
            )
            return

        super().__setattr__(name, value)

    async def generate(self, nodes: list[str] | None = None) -> None:
        self._validate_artifact_definition_support(ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE)

        nodes = nodes or []
        payload = {"nodes": nodes}
        resp = await self._client._post(f"{self._client.address}/api/artifact/generate/{self.id}", payload=payload)
        resp.raise_for_status()

    async def artifact_generate(self, name: str) -> None:
        self._validate_artifact_support(ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE)

        artifact = await self._client.get(kind="CoreArtifact", name__value=name, object__ids=[self.id])
        await artifact._get_relationship_one(name="definition").fetch()
        await artifact._get_relationship_one(name="definition").peer.generate([artifact.id])

    async def artifact_fetch(self, name: str) -> str | dict[str, Any]:
        self._validate_artifact_support(ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE)

        artifact = await self._client.get(kind="CoreArtifact", name__value=name, object__ids=[self.id])
        return await self._client.object_store.get(identifier=artifact._get_attribute(name="storage_id").value)

    async def delete(self, timeout: int | None = None, request_context: RequestContext | None = None) -> None:
        input_data = {"data": {"id": self.id}}
        if context_data := self._get_request_context(request_context=request_context):
            input_data["context"] = context_data

        mutation_query = {"ok": None}
        query = Mutation(
            mutation=f"{self._schema.kind}Delete",
            input_data=input_data,
            query=mutation_query,
        )
        await self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            timeout=timeout,
            tracker=f"mutation-{str(self._schema.kind).lower()}-delete",
        )

    async def save(
        self,
        allow_upsert: bool = False,
        update_group_context: bool | None = None,
        timeout: int | None = None,
        request_context: RequestContext | None = None,
    ) -> None:
        if self._existing is False or allow_upsert is True:
            await self.create(allow_upsert=allow_upsert, timeout=timeout, request_context=request_context)
        else:
            await self.update(timeout=timeout, request_context=request_context)

        if update_group_context is None and self._client.mode == InfrahubClientMode.TRACKING:
            update_group_context = True

        if not isinstance(self._schema, GenericSchemaAPI):
            if "CoreGroup" in self._schema.inherit_from:
                await self._client.group_context.add_related_groups(
                    ids=[self.id], update_group_context=update_group_context
                )
            else:
                await self._client.group_context.add_related_nodes(
                    ids=[self.id], update_group_context=update_group_context
                )
        else:
            await self._client.group_context.add_related_nodes(ids=[self.id], update_group_context=update_group_context)

        self._client.store.set(node=self)

    async def generate_query_data(
        self,
        filters: dict[str, Any] | None = None,
        offset: int | None = None,
        limit: int | None = None,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        fragment: bool = False,
        prefetch_relationships: bool = False,
        partial_match: bool = False,
        property: bool = False,
        order: Order | None = None,
    ) -> dict[str, Any | dict]:
        data = self.generate_query_data_init(
            filters=filters,
            offset=offset,
            limit=limit,
            include=include,
            exclude=exclude,
            partial_match=partial_match,
            order=order,
        )
        data["edges"]["node"].update(
            await self.generate_query_data_node(
                include=include,
                exclude=exclude,
                prefetch_relationships=prefetch_relationships,
                inherited=True,
                property=property,
            )
        )

        if isinstance(self._schema, GenericSchemaAPI) and fragment:
            for child in self._schema.used_by:
                child_schema = await self._client.schema.get(kind=child)
                child_node = InfrahubNode(client=self._client, schema=child_schema)

                # Add the attribute and the relationship already part of the parent to the exclude list for the children
                exclude_parent = self._attributes + self._relationships
                _, _, only_in_list2 = compare_lists(list1=include or [], list2=exclude_parent)

                exclude_child = only_in_list2
                if exclude:
                    exclude_child += exclude

                child_data = await child_node.generate_query_data_node(
                    include=include,
                    exclude=exclude_child,
                    prefetch_relationships=prefetch_relationships,
                    inherited=False,
                    insert_alias=True,
                    property=property,
                )

                if child_data:
                    data["edges"]["node"][f"...on {child}"] = child_data

        return {self._schema.kind: data}

    async def generate_query_data_node(
        self,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        inherited: bool = True,
        insert_alias: bool = False,
        prefetch_relationships: bool = False,
        property: bool = False,
    ) -> dict[str, Any | dict]:
        """Generate the node part of a GraphQL Query with attributes and nodes.

        Args:
            include (Optional[list[str]], optional): List of attributes or relationships to include. Defaults to None.
            exclude (Optional[list[str]], optional): List of attributes or relationships to exclude. Defaults to None.
            inherited (bool, optional): Indicated of the attributes and the relationships inherited from generics should be included as well.
                                        Defaults to True.
            insert_alias (bool, optional): If True, inserts aliases in the query for each attribute or relationship.
            prefetch_relationships (bool, optional): If True, pre-fetches relationship data as part of the query.

        Returns:
            dict[str, Union[Any, Dict]]: GraphQL query in dictionary format
        """

        data: dict[str, Any] = {}

        for attr_name in self._attributes:
            if exclude and attr_name in exclude:
                continue

            attr: Attribute = getattr(self, attr_name)

            if not inherited and attr._schema.inherited:
                continue

            attr_data = attr._generate_query_data(property=property)
            if attr_data:
                data[attr_name] = attr_data
                if insert_alias:
                    data[attr_name]["@alias"] = f"__alias__{self._schema.kind}__{attr_name}"
            elif insert_alias:
                if insert_alias:
                    data[attr_name] = {"@alias": f"__alias__{self._schema.kind}__{attr_name}"}

        for rel_name in self._relationships:
            if exclude and rel_name in exclude:
                continue

            rel_schema = self._schema.get_relationship(name=rel_name)

            if not rel_schema or (not inherited and rel_schema.inherited):
                continue

            if (
                rel_schema.cardinality == RelationshipCardinality.MANY  # type: ignore[union-attr]
                and rel_schema.kind not in [RelationshipKind.ATTRIBUTE, RelationshipKind.PARENT]  # type: ignore[union-attr]
                and not (include and rel_name in include)
            ):
                continue

            peer_data: dict[str, Any] = {}
            should_fetch_relationship = prefetch_relationships or (include is not None and rel_name in include)
            if rel_schema and should_fetch_relationship:
                peer_schema = await self._client.schema.get(kind=rel_schema.peer, branch=self._branch)
                peer_node = InfrahubNode(client=self._client, schema=peer_schema, branch=self._branch)
                peer_data = await peer_node.generate_query_data_node(
                    property=property,
                )

            if rel_schema and rel_schema.cardinality == "one":
                rel_data = RelatedNode._generate_query_data(peer_data=peer_data, property=property)
                # Nodes involved in a hierarchy are required to inherit from a common ancestor node, and graphql
                # tries to resolve attributes in this ancestor instead of actual node. To avoid
                # invalid queries issues when attribute is missing in the common ancestor, we use a fragment
                # to explicit actual node kind we are querying.
                if rel_schema.kind == RelationshipKind.HIERARCHY:
                    data_node = rel_data["node"]
                    rel_data["node"] = {}
                    rel_data["node"][f"...on {rel_schema.peer}"] = data_node
            elif rel_schema and rel_schema.cardinality == "many":
                rel_data = RelationshipManager._generate_query_data(peer_data=peer_data, property=property)

            data[rel_name] = rel_data

            if insert_alias:
                data[rel_name]["@alias"] = f"__alias__{self._schema.kind}__{rel_name}"

        return data

    async def add_relationships(self, relation_to_update: str, related_nodes: list[str]) -> None:
        query = self._relationship_mutation(
            action="Add", relation_to_update=relation_to_update, related_nodes=related_nodes
        )
        tracker = f"mutation-{str(self._schema.kind).lower()}-relationshipadd-{relation_to_update}"
        await self._client.execute_graphql(query=query, branch_name=self._branch, tracker=tracker)

    async def remove_relationships(self, relation_to_update: str, related_nodes: list[str]) -> None:
        query = self._relationship_mutation(
            action="Remove", relation_to_update=relation_to_update, related_nodes=related_nodes
        )
        tracker = f"mutation-{str(self._schema.kind).lower()}-relationshipremove-{relation_to_update}"
        await self._client.execute_graphql(query=query, branch_name=self._branch, tracker=tracker)

    def _generate_mutation_query(self) -> dict[str, Any]:
        query_result: dict[str, Any] = {"ok": None, "object": {"id": None}}

        for attr_name in self._attributes:
            attr: Attribute = getattr(self, attr_name)
            query_result["object"].update(attr._generate_mutation_query())

        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if not isinstance(rel, RelatedNode):
                continue

            query_result["object"].update(rel._generate_mutation_query())

        return query_result

    async def _process_mutation_result(
        self, mutation_name: str, response: dict[str, Any], timeout: int | None = None
    ) -> None:
        object_response: dict[str, Any] = response[mutation_name]["object"]
        self.id = object_response["id"]
        self._existing = True

        for attr_name in self._attributes:
            attr = getattr(self, attr_name)
            if (
                attr_name not in object_response
                or not isinstance(attr.value, InfrahubNodeBase)
                or not attr.value.is_resource_pool()
            ):
                continue

            # Process allocated resource from a pool and update attribute
            attr.value = object_response[attr_name]["value"]

        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if rel_name not in object_response or not isinstance(rel, RelatedNode) or not rel.is_resource_pool:
                continue

            # Process allocated resource from a pool and update related node
            allocated_resource = object_response[rel_name]
            related_node = RelatedNode(
                client=self._client, branch=self._branch, schema=rel.schema, data=allocated_resource
            )
            await related_node.fetch(timeout=timeout)
            setattr(self, rel_name, related_node)

    async def create(
        self, allow_upsert: bool = False, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None:
        mutation_query = self._generate_mutation_query()

        # Upserting means we may want to create, meaning payload contains all mandatory fields required for a creation,
        # so hfid is just redondant information. Currently, upsert mutation has performance overhead if `hfid` is filled.
        if allow_upsert:
            input_data = self._generate_input_data(exclude_hfid=True, request_context=request_context)
            mutation_name = f"{self._schema.kind}Upsert"
            tracker = f"mutation-{str(self._schema.kind).lower()}-upsert"
        else:
            input_data = self._generate_input_data(exclude_hfid=True, request_context=request_context)
            mutation_name = f"{self._schema.kind}Create"
            tracker = f"mutation-{str(self._schema.kind).lower()}-create"
        query = Mutation(
            mutation=mutation_name,
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )
        response = await self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            tracker=tracker,
            variables=input_data["variables"],
            timeout=timeout,
        )
        await self._process_mutation_result(mutation_name=mutation_name, response=response, timeout=timeout)

    async def update(
        self, do_full_update: bool = False, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None:
        input_data = self._generate_input_data(exclude_unmodified=not do_full_update, request_context=request_context)
        mutation_query = self._generate_mutation_query()
        mutation_name = f"{self._schema.kind}Update"

        query = Mutation(
            mutation=mutation_name,
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )
        response = await self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            timeout=timeout,
            tracker=f"mutation-{str(self._schema.kind).lower()}-update",
            variables=input_data["variables"],
        )
        await self._process_mutation_result(mutation_name=mutation_name, response=response, timeout=timeout)

    async def _process_relationships(
        self,
        node_data: dict[str, Any],
        branch: str,
        related_nodes: list[InfrahubNode],
        timeout: int | None = None,
    ) -> None:
        """Processes the Relationships of a InfrahubNode and add Related Nodes to a list.

        Args:
            node_data (dict[str, Any]): The item from the GraphQL response corresponding to the node.
            branch (str): The branch name.
            related_nodes (list[InfrahubNode]): The list to which related nodes will be appended.
            timeout (int, optional): Overrides default timeout used when querying the graphql API. Specified in seconds.
        """
        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if rel and isinstance(rel, RelatedNode):
                relation = node_data["node"].get(rel_name, None)
                if relation and relation.get("node", None):
                    related_node = await InfrahubNode.from_graphql(
                        client=self._client, branch=branch, data=relation, timeout=timeout
                    )
                    related_nodes.append(related_node)
            elif rel and isinstance(rel, RelationshipManager):
                peers = node_data["node"].get(rel_name, None)
                if peers and peers["edges"]:
                    for peer in peers["edges"]:
                        related_node = await InfrahubNode.from_graphql(
                            client=self._client, branch=branch, data=peer, timeout=timeout
                        )
                        related_nodes.append(related_node)

    async def get_pool_allocated_resources(self, resource: InfrahubNode) -> list[InfrahubNode]:
        """Fetch all nodes that were allocated for the pool and a given resource.

        Args:
            resource (InfrahubNode): The resource from which the nodes were allocated.

        Returns:
            list[InfrahubNode]: The allocated nodes.
        """
        if not self.is_resource_pool():
            raise ValueError("Allocated resources can only be fetched from resource pool nodes.")

        graphql_query_name = "InfrahubResourcePoolAllocated"
        node_ids_per_kind: dict[str, list[str]] = {}

        has_remaining_items = True
        page_number = 1
        while has_remaining_items:
            page_offset = (page_number - 1) * self._client.pagination_size

            query = Query(
                query={
                    graphql_query_name: {
                        "@filters": {
                            "pool_id": "$pool_id",
                            "resource_id": "$resource_id",
                            "offset": page_offset,
                            "limit": self._client.pagination_size,
                        },
                        "count": None,
                        "edges": {"node": {"id": None, "kind": None, "branch": None, "identifier": None}},
                    }
                },
                name="GetAllocatedResourceForPool",
                variables={"pool_id": str, "resource_id": str},
            )
            response = await self._client.execute_graphql(
                query=query.render(),
                variables={"pool_id": self.id, "resource_id": resource.id},
                branch_name=self._branch,
                tracker=f"get-allocated-resources-page{page_number}",
            )

            for edge in response[graphql_query_name]["edges"]:
                node = edge["node"]
                node_ids_per_kind.setdefault(node["kind"], []).append(node["id"])

            remaining_items = response[graphql_query_name].get("count", 0) - (
                page_offset + self._client.pagination_size
            )
            if remaining_items <= 0:
                has_remaining_items = False

            page_number += 1

        nodes: list[InfrahubNode] = []
        for kind, node_ids in node_ids_per_kind.items():
            nodes.extend(await self._client.filters(kind=kind, branch=self._branch, ids=node_ids))

        return nodes

    async def get_pool_resources_utilization(self) -> list[dict[str, Any]]:
        """Fetch the utilization of each resource for the pool.

        Returns:
            list[dict[str, Any]]: A list containing the allocation numbers for each resource of the pool.
        """
        if not self.is_resource_pool():
            raise ValueError("Pool utilization can only be fetched for resource pool nodes.")

        graphql_query_name = "InfrahubResourcePoolUtilization"

        query = Query(
            query={
                graphql_query_name: {
                    "@filters": {"pool_id": "$pool_id"},
                    "count": None,
                    "edges": {
                        "node": {
                            "id": None,
                            "kind": None,
                            "utilization": None,
                            "utilization_branches": None,
                            "utilization_default_branch": None,
                        }
                    },
                }
            },
            name="GetUtilizationForPool",
            variables={"pool_id": str},
        )
        response = await self._client.execute_graphql(
            query=query.render(),
            variables={"pool_id": self.id},
            branch_name=self._branch,
            tracker="get-pool-utilization",
        )

        if response[graphql_query_name].get("count", 0):
            return [edge["node"] for edge in response[graphql_query_name]["edges"]]
        return []

    def _get_relationship_many(self, name: str) -> RelationshipManager:
        if name in self._relationship_cardinality_many_data:
            return self._relationship_cardinality_many_data[name]

        raise ResourceNotDefinedError(message=f"The node doesn't have a cardinality=many relationship for {name}")

    def _get_relationship_one(self, name: str) -> RelatedNode:
        if name in self._relationship_cardinality_one_data:
            return self._relationship_cardinality_one_data[name]

        raise ResourceNotDefinedError(message=f"The node doesn't have a cardinality=one relationship for {name}")

    async def get_flat_value(self, key: str, separator: str = "__") -> Any:
        """Query recursively a value defined in a flat notation (string), on a hierarchy of objects

        Examples:
            name__value
            module.object.value
        """
        if separator not in key:
            return getattr(self, key)

        first, remaining = key.split(separator, maxsplit=1)

        if first in self._schema.attribute_names:
            attr = getattr(self, first)
            for part in remaining.split(separator):
                attr = getattr(attr, part)
            return attr

        try:
            rel = self._schema.get_relationship(name=first)
        except ValueError as exc:
            raise ValueError(f"No attribute or relationship named '{first}' for '{self._schema.kind}'") from exc

        if rel.cardinality != RelationshipCardinality.ONE:
            raise ValueError(
                f"Can only look up flat value for relationships of cardinality {RelationshipCardinality.ONE.value}"
            )

        related_node: RelatedNode = getattr(self, first)
        await related_node.fetch()
        return await related_node.peer.get_flat_value(key=remaining, separator=separator)

    async def extract(self, params: dict[str, str]) -> dict[str, Any]:
        """Extract some datapoints defined in a flat notation."""
        result: dict[str, Any] = {}
        for key, value in params.items():
            result[key] = await self.get_flat_value(key=value)

        return result

    def __dir__(self) -> Iterable[str]:
        base = list(super().__dir__())
        return sorted(
            base
            + list(self._attribute_data.keys())
            + list(self._relationship_cardinality_many_data.keys())
            + list(self._relationship_cardinality_one_data.keys())
        )


class InfrahubNodeSync(InfrahubNodeBase):
    """Represents a Infrahub node in a synchronous context."""

    def __init__(
        self,
        client: InfrahubClientSync,
        schema: MainSchemaTypesAPI,
        branch: str | None = None,
        data: dict | None = None,
    ) -> None:
        """
        Args:
            client (InfrahubClientSync): The client used to interact with the backend synchronously.
            schema (MainSchemaTypes): The schema of the node.
            branch (Optional[str]): The branch where the node resides.
            data (Optional[dict]): Optional data to initialize the node.
        """
        self._client = client

        if isinstance(data, dict) and isinstance(data.get("node"), dict):
            data = data.get("node")

        self._relationship_cardinality_many_data: dict[str, RelationshipManagerSync] = {}
        self._relationship_cardinality_one_data: dict[str, RelatedNodeSync] = {}

        super().__init__(schema=schema, branch=branch or client.default_branch, data=data)

    @classmethod
    def from_graphql(
        cls,
        client: InfrahubClientSync,
        branch: str,
        data: dict,
        schema: MainSchemaTypesAPI | None = None,
        timeout: int | None = None,
    ) -> Self:
        if not schema:
            node_kind = data.get("__typename") or data.get("node", {}).get("__typename", None)
            if not node_kind:
                raise ValueError("Unable to determine the type of the node, __typename not present in data")
            schema = client.schema.get(kind=node_kind, branch=branch, timeout=timeout)

        return cls(client=client, schema=schema, branch=branch, data=cls._strip_alias(data))

    def _init_relationships(self, data: dict | None = None) -> None:
        for rel_schema in self._schema.relationships:
            rel_data = data.get(rel_schema.name, None) if isinstance(data, dict) else None

            if rel_schema.cardinality == "one":
                if isinstance(rel_data, RelatedNodeSync):
                    peer_id_data: dict[str, Any] = {
                        key: value
                        for key, value in (
                            ("id", rel_data.id),
                            ("hfid", rel_data.hfid),
                            ("__typename", rel_data.typename),
                            ("kind", rel_data.kind),
                            ("display_label", rel_data.display_label),
                        )
                        if value is not None
                    }
                    if peer_id_data:
                        rel_data = peer_id_data
                    else:
                        rel_data = None
                self._relationship_cardinality_one_data[rel_schema.name] = RelatedNodeSync(
                    name=rel_schema.name, branch=self._branch, client=self._client, schema=rel_schema, data=rel_data
                )
            else:
                self._relationship_cardinality_many_data[rel_schema.name] = RelationshipManagerSync(
                    name=rel_schema.name,
                    client=self._client,
                    node=self,
                    branch=self._branch,
                    schema=rel_schema,
                    data=rel_data,
                )

    def __getattr__(self, name: str) -> Attribute | RelationshipManagerSync | RelatedNodeSync:
        if "_attribute_data" in self.__dict__ and name in self._attribute_data:
            return self._attribute_data[name]
        if "_relationship_cardinality_many_data" in self.__dict__ and name in self._relationship_cardinality_many_data:
            return self._relationship_cardinality_many_data[name]
        if "_relationship_cardinality_one_data" in self.__dict__ and name in self._relationship_cardinality_one_data:
            return self._relationship_cardinality_one_data[name]

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        """Set values for relationship names that exist or revert to normal behaviour"""
        if "_relationship_cardinality_one_data" in self.__dict__ and name in self._relationship_cardinality_one_data:
            rel_schemas = [rel_schema for rel_schema in self._schema.relationships if rel_schema.name == name]
            if not rel_schemas:
                raise SchemaNotFoundError(
                    identifier=self._schema.kind,
                    message=f"Unable to find relationship schema for '{name}' on {self._schema.kind}",
                )
            rel_schema = rel_schemas[0]
            self._relationship_cardinality_one_data[name] = RelatedNodeSync(
                name=rel_schema.name, branch=self._branch, client=self._client, schema=rel_schema, data=value
            )
            return

        super().__setattr__(name, value)

    def generate(self, nodes: list[str] | None = None) -> None:
        self._validate_artifact_definition_support(ARTIFACT_DEFINITION_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE)
        nodes = nodes or []
        payload = {"nodes": nodes}
        resp = self._client._post(f"{self._client.address}/api/artifact/generate/{self.id}", payload=payload)
        resp.raise_for_status()

    def artifact_generate(self, name: str) -> None:
        self._validate_artifact_support(ARTIFACT_GENERATE_FEATURE_NOT_SUPPORTED_MESSAGE)
        artifact = self._client.get(kind="CoreArtifact", name__value=name, object__ids=[self.id])
        artifact._get_relationship_one(name="definition").fetch()
        artifact._get_relationship_one(name="definition").peer.generate([artifact.id])

    def artifact_fetch(self, name: str) -> str | dict[str, Any]:
        self._validate_artifact_support(ARTIFACT_FETCH_FEATURE_NOT_SUPPORTED_MESSAGE)
        artifact = self._client.get(kind="CoreArtifact", name__value=name, object__ids=[self.id])
        return self._client.object_store.get(identifier=artifact._get_attribute(name="storage_id").value)

    def delete(self, timeout: int | None = None, request_context: RequestContext | None = None) -> None:
        input_data = {"data": {"id": self.id}}
        if context_data := self._get_request_context(request_context=request_context):
            input_data["context"] = context_data

        mutation_query = {"ok": None}
        query = Mutation(
            mutation=f"{self._schema.kind}Delete",
            input_data=input_data,
            query=mutation_query,
        )
        self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            tracker=f"mutation-{str(self._schema.kind).lower()}-delete",
            timeout=timeout,
        )

    def save(
        self,
        allow_upsert: bool = False,
        update_group_context: bool | None = None,
        timeout: int | None = None,
        request_context: RequestContext | None = None,
    ) -> None:
        if self._existing is False or allow_upsert is True:
            self.create(allow_upsert=allow_upsert, timeout=timeout, request_context=request_context)
        else:
            self.update(timeout=timeout, request_context=request_context)

        if update_group_context is None and self._client.mode == InfrahubClientMode.TRACKING:
            update_group_context = True

        if not isinstance(self._schema, GenericSchemaAPI):
            if "CoreGroup" in self._schema.inherit_from:
                self._client.group_context.add_related_groups(ids=[self.id], update_group_context=update_group_context)
            else:
                self._client.group_context.add_related_nodes(ids=[self.id], update_group_context=update_group_context)
        else:
            self._client.group_context.add_related_nodes(ids=[self.id], update_group_context=update_group_context)

        self._client.store.set(node=self)

    def generate_query_data(
        self,
        filters: dict[str, Any] | None = None,
        offset: int | None = None,
        limit: int | None = None,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        fragment: bool = False,
        prefetch_relationships: bool = False,
        partial_match: bool = False,
        property: bool = False,
        order: Order | None = None,
    ) -> dict[str, Any | dict]:
        data = self.generate_query_data_init(
            filters=filters,
            offset=offset,
            limit=limit,
            include=include,
            exclude=exclude,
            partial_match=partial_match,
            order=order,
        )
        data["edges"]["node"].update(
            self.generate_query_data_node(
                include=include,
                exclude=exclude,
                prefetch_relationships=prefetch_relationships,
                inherited=True,
                property=property,
            )
        )

        if isinstance(self._schema, GenericSchemaAPI) and fragment:
            for child in self._schema.used_by:
                child_schema = self._client.schema.get(kind=child)
                child_node = InfrahubNodeSync(client=self._client, schema=child_schema)

                exclude_parent = self._attributes + self._relationships
                _, _, only_in_list2 = compare_lists(list1=include or [], list2=exclude_parent)

                exclude_child = only_in_list2
                if exclude:
                    exclude_child += exclude

                child_data = child_node.generate_query_data_node(
                    include=include,
                    exclude=exclude_child,
                    prefetch_relationships=prefetch_relationships,
                    inherited=False,
                    insert_alias=True,
                    property=property,
                )

                if child_data:
                    data["edges"]["node"][f"...on {child}"] = child_data

        return {self._schema.kind: data}

    def generate_query_data_node(
        self,
        include: list[str] | None = None,
        exclude: list[str] | None = None,
        inherited: bool = True,
        insert_alias: bool = False,
        prefetch_relationships: bool = False,
        property: bool = False,
    ) -> dict[str, Any | dict]:
        """Generate the node part of a GraphQL Query with attributes and nodes.

        Args:
            include (Optional[list[str]], optional): List of attributes or relationships to include. Defaults to None.
            exclude (Optional[list[str]], optional): List of attributes or relationships to exclude. Defaults to None.
            inherited (bool, optional): Indicated of the attributes and the relationships inherited from generics should be included as well.
                                        Defaults to True.
            insert_alias (bool, optional): If True, inserts aliases in the query for each attribute or relationship.
            prefetch_relationships (bool, optional): If True, pre-fetches relationship data as part of the query.

        Returns:
            dict[str, Union[Any, Dict]]: GraphQL query in dictionary format
        """

        data: dict[str, Any] = {}

        for attr_name in self._attributes:
            if exclude and attr_name in exclude:
                continue

            attr: Attribute = getattr(self, attr_name)

            if not inherited and attr._schema.inherited:
                continue

            attr_data = attr._generate_query_data(property=property)
            if attr_data:
                data[attr_name] = attr_data
                if insert_alias:
                    data[attr_name]["@alias"] = f"__alias__{self._schema.kind}__{attr_name}"
            elif insert_alias:
                if insert_alias:
                    data[attr_name] = {"@alias": f"__alias__{self._schema.kind}__{attr_name}"}

        for rel_name in self._relationships:
            if exclude and rel_name in exclude:
                continue

            rel_schema = self._schema.get_relationship(name=rel_name)

            if not rel_schema or (not inherited and rel_schema.inherited):
                continue

            if (
                rel_schema.cardinality == RelationshipCardinality.MANY  # type: ignore[union-attr]
                and rel_schema.kind not in [RelationshipKind.ATTRIBUTE, RelationshipKind.PARENT]  # type: ignore[union-attr]
                and not (include and rel_name in include)
            ):
                continue

            peer_data: dict[str, Any] = {}
            should_fetch_relationship = prefetch_relationships or (include is not None and rel_name in include)
            if rel_schema and should_fetch_relationship:
                peer_schema = self._client.schema.get(kind=rel_schema.peer, branch=self._branch)
                peer_node = InfrahubNodeSync(client=self._client, schema=peer_schema, branch=self._branch)
                peer_data = peer_node.generate_query_data_node(include=include, exclude=exclude, property=property)

            if rel_schema and rel_schema.cardinality == "one":
                rel_data = RelatedNodeSync._generate_query_data(peer_data=peer_data, property=property)
                # Nodes involved in a hierarchy are required to inherit from a common ancestor node, and graphql
                # tries to resolve attributes in this ancestor instead of actual node. To avoid
                # invalid queries issues when attribute is missing in the common ancestor, we use a fragment
                # to explicit actual node kind we are querying.
                if rel_schema.kind == RelationshipKind.HIERARCHY:
                    data_node = rel_data["node"]
                    rel_data["node"] = {}
                    rel_data["node"][f"...on {rel_schema.peer}"] = data_node
            elif rel_schema and rel_schema.cardinality == "many":
                rel_data = RelationshipManagerSync._generate_query_data(peer_data=peer_data, property=property)

            data[rel_name] = rel_data

            if insert_alias:
                data[rel_name]["@alias"] = f"__alias__{self._schema.kind}__{rel_name}"

        return data

    def add_relationships(
        self,
        relation_to_update: str,
        related_nodes: list[str],
    ) -> None:
        query = self._relationship_mutation(
            action="Add", relation_to_update=relation_to_update, related_nodes=related_nodes
        )
        tracker = f"mutation-{str(self._schema.kind).lower()}-relationshipadd-{relation_to_update}"
        self._client.execute_graphql(query=query, branch_name=self._branch, tracker=tracker)

    def remove_relationships(self, relation_to_update: str, related_nodes: list[str]) -> None:
        query = self._relationship_mutation(
            action="Remove", relation_to_update=relation_to_update, related_nodes=related_nodes
        )
        tracker = f"mutation-{str(self._schema.kind).lower()}-relationshipremove-{relation_to_update}"
        self._client.execute_graphql(query=query, branch_name=self._branch, tracker=tracker)

    def _generate_mutation_query(self) -> dict[str, Any]:
        query_result: dict[str, Any] = {"ok": None, "object": {"id": None}}

        for attr_name in self._attributes:
            attr: Attribute = getattr(self, attr_name)
            query_result["object"].update(attr._generate_mutation_query())

        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if not isinstance(rel, RelatedNodeSync):
                continue

            query_result["object"].update(rel._generate_mutation_query())

        return query_result

    def _process_mutation_result(
        self, mutation_name: str, response: dict[str, Any], timeout: int | None = None
    ) -> None:
        object_response: dict[str, Any] = response[mutation_name]["object"]
        self.id = object_response["id"]
        self._existing = True

        for attr_name in self._attributes:
            attr = getattr(self, attr_name)
            if (
                attr_name not in object_response
                or not isinstance(attr.value, InfrahubNodeBase)
                or not attr.value.is_resource_pool()
            ):
                continue

            # Process allocated resource from a pool and update attribute
            attr.value = object_response[attr_name]["value"]

        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if rel_name not in object_response or not isinstance(rel, RelatedNodeSync) or not rel.is_resource_pool:
                continue

            # Process allocated resource from a pool and update related node
            allocated_resource = object_response[rel_name]
            related_node = RelatedNodeSync(
                client=self._client, branch=self._branch, schema=rel.schema, data=allocated_resource
            )
            related_node.fetch(timeout=timeout)
            setattr(self, rel_name, related_node)

    def create(
        self, allow_upsert: bool = False, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None:
        mutation_query = self._generate_mutation_query()

        if allow_upsert:
            input_data = self._generate_input_data(exclude_hfid=False, request_context=request_context)
            mutation_name = f"{self._schema.kind}Upsert"
            tracker = f"mutation-{str(self._schema.kind).lower()}-upsert"
        else:
            input_data = self._generate_input_data(exclude_hfid=True, request_context=request_context)
            mutation_name = f"{self._schema.kind}Create"
            tracker = f"mutation-{str(self._schema.kind).lower()}-create"
        query = Mutation(
            mutation=mutation_name,
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )

        response = self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            tracker=tracker,
            variables=input_data["variables"],
            timeout=timeout,
        )
        self._process_mutation_result(mutation_name=mutation_name, response=response, timeout=timeout)

    def update(
        self, do_full_update: bool = False, timeout: int | None = None, request_context: RequestContext | None = None
    ) -> None:
        input_data = self._generate_input_data(exclude_unmodified=not do_full_update, request_context=request_context)
        mutation_query = self._generate_mutation_query()
        mutation_name = f"{self._schema.kind}Update"

        query = Mutation(
            mutation=mutation_name,
            input_data=input_data["data"],
            query=mutation_query,
            variables=input_data["mutation_variables"],
        )

        response = self._client.execute_graphql(
            query=query.render(),
            branch_name=self._branch,
            tracker=f"mutation-{str(self._schema.kind).lower()}-update",
            variables=input_data["variables"],
            timeout=timeout,
        )
        self._process_mutation_result(mutation_name=mutation_name, response=response, timeout=timeout)

    def _process_relationships(
        self,
        node_data: dict[str, Any],
        branch: str,
        related_nodes: list[InfrahubNodeSync],
        timeout: int | None = None,
    ) -> None:
        """Processes the Relationships of a InfrahubNodeSync and add Related Nodes to a list.

        Args:
            node_data (dict[str, Any]): The item from the GraphQL response corresponding to the node.
            branch (str): The branch name.
            related_nodes (list[InfrahubNodeSync]): The list to which related nodes will be appended.
            timeout (int, optional): Overrides default timeout used when querying the graphql API. Specified in seconds.

        """
        for rel_name in self._relationships:
            rel = getattr(self, rel_name)
            if rel and isinstance(rel, RelatedNodeSync):
                relation = node_data["node"].get(rel_name, None)
                if relation and relation.get("node", None):
                    related_node = InfrahubNodeSync.from_graphql(
                        client=self._client, branch=branch, data=relation, timeout=timeout
                    )
                    related_nodes.append(related_node)
            elif rel and isinstance(rel, RelationshipManagerSync):
                peers = node_data["node"].get(rel_name, None)
                if peers and peers["edges"]:
                    for peer in peers["edges"]:
                        related_node = InfrahubNodeSync.from_graphql(
                            client=self._client, branch=branch, data=peer, timeout=timeout
                        )
                        related_nodes.append(related_node)

    def get_pool_allocated_resources(self, resource: InfrahubNodeSync) -> list[InfrahubNodeSync]:
        """Fetch all nodes that were allocated for the pool and a given resource.

        Args:
            resource (InfrahubNodeSync): The resource from which the nodes were allocated.

        Returns:
            list[InfrahubNodeSync]: The allocated nodes.
        """
        if not self.is_resource_pool():
            raise ValueError("Allocate resources can only be fetched from resource pool nodes.")

        graphql_query_name = "InfrahubResourcePoolAllocated"
        node_ids_per_kind: dict[str, list[str]] = {}

        has_remaining_items = True
        page_number = 1
        while has_remaining_items:
            page_offset = (page_number - 1) * self._client.pagination_size

            query = Query(
                query={
                    graphql_query_name: {
                        "@filters": {
                            "pool_id": "$pool_id",
                            "resource_id": "$resource_id",
                            "offset": page_offset,
                            "limit": self._client.pagination_size,
                        },
                        "count": None,
                        "edges": {"node": {"id": None, "kind": None, "branch": None, "identifier": None}},
                    }
                },
                name="GetAllocatedResourceForPool",
                variables={"pool_id": str, "resource_id": str},
            )
            response = self._client.execute_graphql(
                query=query.render(),
                variables={"pool_id": self.id, "resource_id": resource.id},
                branch_name=self._branch,
                tracker=f"get-allocated-resources-page{page_number}",
            )

            for edge in response[graphql_query_name]["edges"]:
                node = edge["node"]
                node_ids_per_kind.setdefault(node["kind"], []).append(node["id"])

            remaining_items = response[graphql_query_name].get("count", 0) - (
                page_offset + self._client.pagination_size
            )
            if remaining_items <= 0:
                has_remaining_items = False

            page_number += 1

        nodes: list[InfrahubNodeSync] = []
        for kind, node_ids in node_ids_per_kind.items():
            nodes.extend(self._client.filters(kind=kind, branch=self._branch, ids=node_ids))

        return nodes

    def get_pool_resources_utilization(self) -> list[dict[str, Any]]:
        """Fetch the utilization of each resource for the pool.

        Returns:
            list[dict[str, Any]]: A list containing the allocation numbers for each resource of the pool.
        """
        if not self.is_resource_pool():
            raise ValueError("Pool utilization can only be fetched for resource pool nodes.")

        graphql_query_name = "InfrahubResourcePoolUtilization"

        query = Query(
            query={
                graphql_query_name: {
                    "@filters": {"pool_id": "$pool_id"},
                    "count": None,
                    "edges": {
                        "node": {
                            "id": None,
                            "kind": None,
                            "utilization": None,
                            "utilization_branches": None,
                            "utilization_default_branch": None,
                        }
                    },
                }
            },
            name="GetUtilizationForPool",
            variables={"pool_id": str},
        )
        response = self._client.execute_graphql(
            query=query.render(),
            variables={"pool_id": self.id},
            branch_name=self._branch,
            tracker="get-pool-utilization",
        )

        if response[graphql_query_name].get("count", 0):
            return [edge["node"] for edge in response[graphql_query_name]["edges"]]
        return []

    def _get_relationship_many(self, name: str) -> RelationshipManager | RelationshipManagerSync:
        if name in self._relationship_cardinality_many_data:
            return self._relationship_cardinality_many_data[name]

        raise ResourceNotDefinedError(message=f"The node doesn't have a cardinality=many relationship for {name}")

    def _get_relationship_one(self, name: str) -> RelatedNode | RelatedNodeSync:
        if name in self._relationship_cardinality_one_data:
            return self._relationship_cardinality_one_data[name]

        raise ResourceNotDefinedError(message=f"The node doesn't have a cardinality=one relationship for {name}")

    def get_flat_value(self, key: str, separator: str = "__") -> Any:
        """Query recursively a value defined in a flat notation (string), on a hierarchy of objects

        Examples:
            name__value
            module.object.value
        """
        if separator not in key:
            return getattr(self, key)

        first, remaining = key.split(separator, maxsplit=1)

        if first in self._schema.attribute_names:
            attr = getattr(self, first)
            for part in remaining.split(separator):
                attr = getattr(attr, part)
            return attr

        try:
            rel = self._schema.get_relationship(name=first)
        except ValueError as exc:
            raise ValueError(f"No attribute or relationship named '{first}' for '{self._schema.kind}'") from exc

        if rel.cardinality != RelationshipCardinality.ONE:
            raise ValueError(
                f"Can only look up flat value for relationships of cardinality {RelationshipCardinality.ONE.value}"
            )

        related_node: RelatedNodeSync = getattr(self, first)
        related_node.fetch()
        return related_node.peer.get_flat_value(key=remaining, separator=separator)

    def extract(self, params: dict[str, str]) -> dict[str, Any]:
        """Extract some datapoints defined in a flat notation."""
        result: dict[str, Any] = {}
        for key, value in params.items():
            result[key] = self.get_flat_value(key=value)

        return result

    def __dir__(self) -> Iterable[str]:
        base = list(super().__dir__())
        return sorted(
            base
            + list(self._attribute_data.keys())
            + list(self._relationship_cardinality_many_data.keys())
            + list(self._relationship_cardinality_one_data.keys())
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from pydantic import BaseModel, Field

from infrahub.utils import InfrahubStringEnum

if TYPE_CHECKING:
    from infrahub.core.constants import BranchSupportType, RelationshipStatus

# Root & Branch nodes are described in their own files


class GraphRelDirection(InfrahubStringEnum):
    EITHER = "either"
    OUTBOUND = "outbound"
    INBOUND = "inbound"


class GraphRelationship(BaseModel):
    peer: str
    direction: GraphRelDirection


class GraphVertex(BaseModel):
    default_label: str = "NULL"

    @classmethod
    def get_default_label(cls) -> str:
        return cls.model_fields["default_label"].default


# -----------------------------------------------------
# Node
# -----------------------------------------------------
class GraphNodeProperties(BaseModel):
    branch_support: BranchSupportType = Field(..., description="Type of branch support for the node")
    kind: str = Field(..., description="Kind of the Node")
    namespace: str = Field(
        ...,
        description="Namespace of the Node, already included in the kind "
        "but having it as a separate properties makes it easier to query by namespace",
    )
    uuid: str = Field(..., description="UUID of the node")


class GraphNodeRelationships(BaseModel):
    IS_RELATED: GraphRelationship = Field(
        GraphRelationship(peer="Relationship", direction=GraphRelDirection.EITHER),
        description="Attach a Relationship to a node",
    )
    HAS_ATTRIBUTE: GraphRelationship = Field(
        GraphRelationship(peer="Attribute", direction=GraphRelDirection.OUTBOUND),
        description="Attach an Attribute to a node",
    )
    IS_PART_OF: GraphRelationship = Field(
        GraphRelationship(peer="Root", direction=GraphRelDirection.OUTBOUND),
        description="Attach the node with the Root node",
    )
    # for the properties
    HAS_SOURCE: GraphRelationship = Field(
        GraphRelationship(peer="Attribute|Relationship", direction=GraphRelDirection.INBOUND),
        description="Identify the node as the source of the object on the other side",
    )
    HAS_OWNER: GraphRelationship = Field(
        GraphRelationship(peer="Attribute|Relationship", direction=GraphRelDirection.INBOUND),
        description="Identify the node as the owner of the object on the other side",
    )


class GraphNodeNode(GraphVertex):
    default_label: str = "Node"  # Most node also have CoreNode
    properties: GraphNodeProperties
    relationships: GraphNodeRelationships


# -----------------------------------------------------
# Relationship
# -----------------------------------------------------
class GraphRelationshipProperties(BaseModel):
    branch_support: BranchSupportType = Field(..., description="Type of branch support for the relationship")
    name: str = Field(..., description="identifier of the relationship")
    uuid: str = Field(..., description="UUID of the relationship")


class GraphRelationshipRelationships(BaseModel):
    IS_RELATED: GraphRelationship = Field(
        GraphRelationship(peer="Node", direction=GraphRelDirection.EITHER),
        description="Attach a Relationship to a node",
    )
    HAS_SOURCE: GraphRelationship = Field(
        GraphRelationship(peer="Node", direction=GraphRelDirection.OUTBOUND), description=""
    )
    HAS_OWNER: GraphRelationship = Field(
        GraphRelationship(peer="Node", direction=GraphRelDirection.OUTBOUND), description=""
    )
    IS_VISIBLE: GraphRelationship = Field(
        GraphRelationship(peer="Boolean", direction=GraphRelDirection.OUTBOUND), description=""
    )
    IS_PROTECTED: GraphRelationship = Field(
        GraphRelationship(peer="Boolean", direction=GraphRelDirection.OUTBOUND), description=""
    )


class GraphRelationshipNode(GraphVertex):
    default_label: str = "Relationship"
    properties: GraphRelationshipProperties
    relationships: GraphRelationshipRelationships


# -----------------------------------------------------
# Attribute
# -----------------------------------------------------
class GraphAttributeProperties(BaseModel):
    branch_support: BranchSupportType = Field(..., description="Type of branch support for the attribute")
    name: str = Field(..., description="name of the attribute as defined in the schema")
    uuid: str = Field(..., description="UUID of the attribute, must be unique")


class GraphAttributeRelationships(BaseModel):
    HAS_ATTRIBUTE: GraphRelationship = Field(
        GraphRelationship(peer="Node", direction=GraphRelDirection.INBOUND),
        description="Identify the node for this Attribute",
    )
    HAS_VALUE: GraphRelationship = Field(
        GraphRelationship(peer="AttributeValue", direction=GraphRelDirection.OUTBOUND),
        description="Attach a Value to an Attribute",
    )
    HAS_SOURCE: GraphRelationship = Field(
        GraphRelationship(peer="Node", direction=GraphRelDirection.OUTBOUND), description=""
    )
    HAS_OWNER: GraphRelationship = Field(
        GraphRelationship(peer="Node", direction=GraphRelDirection.OUTBOUND), description=""
    )
    IS_VISIBLE: GraphRelationship = Field(
        GraphRelationship(peer="Boolean", direction=GraphRelDirection.OUTBOUND), description=""
    )
    IS_PROTECTED: GraphRelationship = Field(
        GraphRelationship(peer="Boolean", direction=GraphRelDirection.OUTBOUND), description=""
    )


class GraphAttributeNode(GraphVertex):
    default_label: str = "Attribute"
    properties: GraphAttributeProperties
    relationships: GraphAttributeRelationships


# -----------------------------------------------------
# AttributeValue
# -----------------------------------------------------
class GraphAttributeValueProperties(BaseModel):
    value: Any = Field(..., description="value of the attribute")


class GraphAttributeIPNetworkProperties(BaseModel):
    value: str = Field(..., description="value of the attribute")
    is_default: bool = Field(..., description="Flag to indicate if an attribute has the default value")
    binary_address: str = Field(..., description="Network address represented in binary format")
    version: int = Field(..., description="Version of IP, either 4 or 6")
    # num_addresses: int = Field(..., description="Total number of addresses available in this IPNetwork")


class GraphAttributeIPHostProperties(BaseModel):
    value: str = Field(..., description="value of the attribute")
    is_default: bool = Field(..., description="Flag to indicate if an attribute has the default value")
    binary_address: str = Field(..., description="Network address represented in binary format")
    version: int = Field(..., description="Version of IP, either 4 or 6")


class GraphAttributeValueRelationships(BaseModel):
    HAS_VALUE: GraphRelationship = Field(
        GraphRelationship(peer="Attribute", direction=GraphRelDirection.INBOUND),
        description="Attach the Value to an Attribute",
    )


class GraphAttributeValueNode(GraphVertex):
    default_label: str = "AttributeValue"
    properties: GraphAttributeValueProperties
    relationships: GraphAttributeValueRelationships


class GraphAttributeValueIndexedNode(GraphVertex):
    default_label: str = "AttributeValueIndexed"
    properties: GraphAttributeValueProperties
    relationships: GraphAttributeValueRelationships


class GraphAttributeIPNetworkNode(GraphVertex):
    default_label: str = "AttributeIPNetwork"
    properties: GraphAttributeIPNetworkProperties
    relationships: GraphAttributeValueRelationships


class GraphAttributeIPHostNode(GraphVertex):
    default_label: str = "AttributeIPHost"
    properties: GraphAttributeIPHostProperties
    relationships: GraphAttributeValueRelationships


# -----------------------------------------------------
# Boolean
# -----------------------------------------------------
class GraphBooleanProperties(BaseModel):
    value: bool = Field(..., description="Value of the Boolean")


class GraphBooleanRelationships(BaseModel):
    IS_VISIBLE: GraphRelationship = Field(
        GraphRelationship(peer="Attribute|Relationship", direction=GraphRelDirection.INBOUND), description=""
    )
    IS_PROTECTED: GraphRelationship = Field(
        GraphRelationship(peer="Attribute|Relationship", direction=GraphRelDirection.INBOUND), description=""
    )


class GraphBooleanNode(GraphVertex):
    default_label: str = "Boolean"
    properties: GraphBooleanProperties
    relationships: GraphBooleanRelationships


class GraphRelationshipIsPartOf(BaseModel):
    from_: str = Field(..., description="Time from which the relationship is valid", alias="from")
    to_: Optional[str] = Field(None, description="Time until which the relationship is valid", alias="to")
    status: RelationshipStatus = Field(..., description="status of the relationship")


class GraphRelationshipDefault(BaseModel):
    branch: str = Field(
        ..., description="name of the branch this relationship is part of, global branch will be -global-"
    )
    branch_level: int = Field(
        ...,
        description="Indicator of the level of the branch compared to main, currently either 1 or 2 since we only support 1 level",
        ge=1,
    )
    from_: str = Field(..., description="Time from which the relationship is valid", alias="from")
    to_: Optional[str] = Field(None, description="Time until which the relationship is valid", alias="to")
    status: RelationshipStatus = Field(..., description="status of the relationship")
    hierarchy: Optional[str] = Field(None, description="Name of the hierarchy this relationship is part of")


def get_graph_schema() -> dict[str, dict[str, Any]]:
    """Generate the graph schema dictionary containing nodes and relationships."""
    graph_node_list: list[type[GraphVertex]] = [
        GraphNodeNode,
        GraphRelationshipNode,
        GraphAttributeNode,
        GraphAttributeValueNode,
        GraphAttributeValueIndexedNode,
        GraphAttributeIPNetworkNode,
        GraphAttributeIPHostNode,
        GraphBooleanNode,
    ]
    return {
        "nodes": {graph_node.get_default_label(): graph_node for graph_node in graph_node_list},
        "relationships": {
            # Ignoring IS_PART_OF for now, because there is a bit of cleanup required
            # "IS_PART_OF": GraphRelationshipIsPartOf,
            "HAS_VALUE": GraphRelationshipDefault,
            "HAS_ATTRIBUTE": GraphRelationshipDefault,
            "IS_RELATED": GraphRelationshipDefault,
            "HAS_SOURCE": GraphRelationshipDefault,
            "HAS_OWNER": GraphRelationshipDefault,
            "IS_VISIBLE": GraphRelationshipDefault,
            "IS_PROTECTED": GraphRelationshipDefault,
        },
    }


GRAPH_SCHEMA: dict[str, dict[str, Any]] = get_graph_schema()



import asyncio
from typing import Any, Coroutine

from infrahub_sdk.protocols import CoreValidator

from infrahub.context import InfrahubContext
from infrahub.core.constants import ValidatorConclusion, ValidatorState
from infrahub.core.timestamp import Timestamp
from infrahub.services.adapters.event import InfrahubEventService
from infrahub.validators.events import send_failed_validator, send_passed_validator


async def run_checks_and_update_validator(
    checks: list[Coroutine[Any, None, ValidatorConclusion]],
    validator: CoreValidator,
    context: InfrahubContext,
    event_service: InfrahubEventService,
    proposed_change_id: str,
) -> None:
    """
    Execute a list of checks coroutines, and set validator fields accordingly.
    Tasks are retrieved by completion order so as soon as we detect a failing check,
    we set validator conclusion to failure.
    """

    # First set validator to in progress, then wait for results
    validator.state.value = ValidatorState.IN_PROGRESS.value
    validator.started_at.value = Timestamp().to_string()
    validator.completed_at.value = ""
    await validator.save()

    failed_early = False

    for earliest_task in asyncio.as_completed(checks):
        result = await earliest_task
        if validator.conclusion.value != ValidatorConclusion.FAILURE.value and result == ValidatorConclusion.FAILURE:
            validator.conclusion.value = ValidatorConclusion.FAILURE.value
            failed_early = True
            await validator.save()
            await send_failed_validator(
                event_service=event_service, validator=validator, proposed_change_id=proposed_change_id, context=context
            )
            # Continue to iterate to wait for the end of all checks

    validator.state.value = ValidatorState.COMPLETED.value
    validator.completed_at.value = Timestamp().to_string()
    if validator.conclusion.value != ValidatorConclusion.FAILURE.value:
        validator.conclusion.value = ValidatorConclusion.SUCCESS.value

    await validator.save()

    if not failed_early:
        if validator.conclusion.value == ValidatorConclusion.SUCCESS.value:
            await send_passed_validator(
                event_service=event_service, validator=validator, proposed_change_id=proposed_change_id, context=context
            )
        else:
            await send_failed_validator(
                event_service=event_service, validator=validator, proposed_change_id=proposed_change_id, context=context
            )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from rich.progress import Progress

from infrahub.core.branch.models import Branch
from infrahub.core.initialization import get_root_node
from infrahub.core.manager import NodeManager
from infrahub.core.migrations.shared import MigrationResult, get_migration_console
from infrahub.core.query import Query, QueryType
from infrahub.core.timestamp import Timestamp
from infrahub.log import get_logger
from infrahub.profiles.node_applier import NodeProfilesApplier

from ..shared import MigrationRequiringRebase
from .load_schema_branch import get_or_load_schema_branch

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase

log = get_logger()


class GetUpdatedProfilesForBranchQuery(Query):
    """
    Get CoreProfile UUIDs with updated attributes on this branch
    """

    name = "get_profiles_by_branch"
    type = QueryType.READ

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["branch"] = self.branch.name
        query = """
MATCH (profile:CoreProfile)-[:HAS_ATTRIBUTE]->(attr:Attribute)-[e:HAS_VALUE]->(:AttributeValue)
WHERE e.branch = $branch
WITH DISTINCT profile.uuid AS profile_uuid
        """
        self.add_to_query(query)
        self.return_labels = ["profile_uuid"]

    def get_profile_ids(self) -> list[str]:
        """Get list of updated profile UUIDs"""
        return [result.get_as_type("profile_uuid", str) for result in self.get_results()]


class GetNodesWithProfileUpdatesForBranchQuery(Query):
    """
    Get Node UUIDs by which branches they have updated profiles on
    """

    name = "get_nodes_with_profile_updates_by_branch"
    type = QueryType.READ

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["branch"] = self.branch.name
        query = """
MATCH (node:Node)-[e:IS_RELATED]->(:Relationship {name: "node__profile"})
WHERE NOT node:CoreProfile
AND e.branch = $branch
WITH DISTINCT node.uuid AS node_uuid
        """
        self.add_to_query(query)
        self.return_labels = ["node_uuid"]

    def get_node_ids(self) -> list[str]:
        """Get list of updated node UUIDs"""
        return [result.get_as_type("node_uuid", str) for result in self.get_results()]


class Migration042(MigrationRequiringRebase):
    """
    Save profile attribute values on each node using the profile in the database
    For any profile that has updates on a given branch (including default branch)
    - run NodeProfilesApplier.apply_profiles on each node related to the profile on that branch
    For any node that has an updated relationship to a profile on a given branch
    - run NodeProfilesApplier.apply_profiles on the node on that branch
    """

    name: str = "042_profile_attrs_in_db"
    minimum_version: int = 41

    def _get_profile_applier(self, db: InfrahubDatabase, branch: Branch) -> NodeProfilesApplier:
        return NodeProfilesApplier(db=db, branch=branch)

    async def validate_migration(self, db: InfrahubDatabase) -> MigrationResult:  # noqa: ARG002
        return MigrationResult()

    async def execute(self, db: InfrahubDatabase) -> MigrationResult:
        root_node = await get_root_node(db=db, initialize=False)
        default_branch_name = root_node.default_branch
        default_branch = await Branch.get_by_name(db=db, name=default_branch_name)
        return await self._do_execute_for_branch(db=db, branch=default_branch)

    async def execute_against_branch(self, db: InfrahubDatabase, branch: Branch) -> MigrationResult:
        return await self._do_execute_for_branch(db=db, branch=branch)

    async def _do_execute_for_branch(self, db: InfrahubDatabase, branch: Branch) -> MigrationResult:
        console = get_migration_console()
        result = MigrationResult()
        await get_or_load_schema_branch(db=db, branch=branch)

        console.print(f"Gathering profiles for branch {branch.name}...", end="")
        get_updated_profiles_for_branch_query = await GetUpdatedProfilesForBranchQuery.init(db=db, branch=branch)
        await get_updated_profiles_for_branch_query.execute(db=db)
        profile_ids = get_updated_profiles_for_branch_query.get_profile_ids()

        profiles_map = await NodeManager.get_many(db=db, branch=branch, ids=list(profile_ids))
        console.print("done")

        node_ids_to_update: set[str] = set()
        with Progress(console=console) as progress:
            gather_nodes_task = progress.add_task(
                f"Gathering affected objects for each profile on branch {branch.name}...", total=len(profiles_map)
            )

            for profile in profiles_map.values():
                node_relationship_manager = profile.get_relationship("related_nodes")
                node_peers = await node_relationship_manager.get_db_peers(db=db)
                node_ids_to_update.update(str(peer.peer_id) for peer in node_peers)
                progress.update(gather_nodes_task, advance=1)
        console.log(f"Collected nodes impacted by profiles on branch {branch.name}.")

        console.print("Identifying nodes with profile updates by branch...", end="")
        get_nodes_with_profile_updates_by_branch_query = await GetNodesWithProfileUpdatesForBranchQuery.init(
            db=db, branch=branch
        )
        await get_nodes_with_profile_updates_by_branch_query.execute(db=db)
        node_ids_to_update.update(get_nodes_with_profile_updates_by_branch_query.get_node_ids())
        console.print("done")

        right_now = Timestamp()
        console.log("Applying profiles to nodes...")
        with Progress(console=console) as progress:
            apply_task = progress.add_task("Applying profiles to nodes...", total=len(node_ids_to_update))
            applier = self._get_profile_applier(db=db, branch=branch)
            for node_id in node_ids_to_update:
                node = await NodeManager.get_one(db=db, branch=branch, id=node_id, at=right_now)
                if node:
                    updated_field_names = await applier.apply_profiles(node=node)
                    if updated_field_names:
                        await node.save(db=db, fields=updated_field_names, at=right_now)
                progress.update(apply_task, advance=1)
        console.log("Completed applying profiles to nodes.")

        return result



from abc import ABC, abstractmethod
from typing import Any


class DataProcessor(ABC):
    """Abstract base class for data processing strategies"""

    @abstractmethod
    async def process_data(self, data: list[dict[str, Any]]) -> list[dict[str, Any]]:
        """Process the data according to the strategy"""



from __future__ import annotations

from typing import TYPE_CHECKING

from infrahub.core import registry
from infrahub.core.constants import NULL_VALUE
from infrahub.core.schema.basenode_schema import (
    UniquenessConstraintType,
    UniquenessConstraintViolation,
)
from infrahub.core.validators.uniqueness.model import (
    NodeUniquenessQueryRequestValued,
    QueryAttributePathValued,
    QueryRelationshipPathValued,
)
from infrahub.core.validators.uniqueness.query import UniquenessValidationQuery
from infrahub.exceptions import HFIDViolatedError, ValidationError

from .interface import NodeConstraintInterface

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.core.relationship.model import RelationshipManager
    from infrahub.core.schema import (
        MainSchemaTypes,
        SchemaAttributePath,
    )
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


class NodeGroupedUniquenessConstraint(NodeConstraintInterface):
    def __init__(self, db: InfrahubDatabase, branch: Branch) -> None:
        self.db = db
        self.branch = branch
        self.schema_branch = registry.schema.get_schema_branch(branch.name)

    async def _get_unique_valued_paths(
        self,
        updated_node: Node,
        path_group: list[SchemaAttributePath],
        filters: list[str],
    ) -> list[QueryAttributePathValued | QueryRelationshipPathValued]:
        # if filters are provided, we need to check if the path group is relevant to the filters
        if filters:
            field_names: list[str] = []
            for schema_attribute_path in path_group:
                if schema_attribute_path.relationship_schema:
                    field_names.append(schema_attribute_path.relationship_schema.name)
                elif schema_attribute_path.attribute_schema:
                    field_names.append(schema_attribute_path.attribute_schema.name)

            if not set(field_names) & set(filters):
                return []

        valued_paths: list[QueryAttributePathValued | QueryRelationshipPathValued] = []
        for schema_attribute_path in path_group:
            if schema_attribute_path.relationship_schema:
                relationship_name = schema_attribute_path.relationship_schema.name
                relationship_manager: RelationshipManager = getattr(updated_node, relationship_name)
                related_node = await relationship_manager.get_peer(db=self.db)
                related_node_id = related_node.get_id() if related_node else None
                valued_paths.append(
                    QueryRelationshipPathValued(
                        relationship_schema=schema_attribute_path.relationship_schema,
                        peer_id=related_node_id,
                        attribute_name=None,
                        attribute_value=None,
                    )
                )
            elif schema_attribute_path.attribute_schema:
                attribute_name = schema_attribute_path.attribute_schema.name
                attribute_field = getattr(updated_node, attribute_name)
                attribute_value = getattr(attribute_field, schema_attribute_path.attribute_property_name or "value")
                if attribute_field.is_enum and attribute_value:
                    attribute_value = attribute_value.value
                elif attribute_value is None:
                    attribute_value = NULL_VALUE
                valued_paths.append(
                    QueryAttributePathValued(
                        attribute_name=attribute_name,
                        value=attribute_value,
                    )
                )
        return valued_paths

    async def _get_single_schema_violations(
        self,
        node: Node,
        node_schema: MainSchemaTypes,
        filters: list[str],
        at: Timestamp | None = None,
    ) -> list[UniquenessConstraintViolation]:
        schema_branch = self.db.schema.get_schema_branch(name=self.branch.name)

        uniqueness_constraint_paths = node_schema.get_unique_constraint_schema_attribute_paths(
            schema_branch=schema_branch
        )

        violations: list[UniquenessConstraintViolation] = []
        for uniqueness_constraint_path in uniqueness_constraint_paths:
            valued_paths = await self._get_unique_valued_paths(
                updated_node=node,
                path_group=uniqueness_constraint_path.attributes_paths,
                filters=filters,
            )

            if not valued_paths:
                continue

            # Create the valued query request for this constraint
            valued_query_request = NodeUniquenessQueryRequestValued(
                kind=node_schema.kind,
                unique_valued_paths=valued_paths,
            )

            # Execute the query
            query = await UniquenessValidationQuery.init(
                db=self.db,
                branch=self.branch,
                at=at,
                query_request=valued_query_request,
                node_ids_to_exclude=[node.get_id()],
            )
            await query.execute(db=self.db)

            # Get violation nodes from the query results
            violation_nodes = query.get_violation_nodes()
            if not violation_nodes:
                continue

            # Create violation object
            uniqueness_constraint_fields = []
            for valued_path in valued_paths:
                if isinstance(valued_path, QueryRelationshipPathValued):
                    uniqueness_constraint_fields.append(valued_path.relationship_schema.name)
                elif isinstance(valued_path, QueryAttributePathValued):
                    uniqueness_constraint_fields.append(valued_path.attribute_name)

            matching_node_ids = {node_id for node_id, _ in violation_nodes}
            if matching_node_ids:
                violations.append(
                    UniquenessConstraintViolation(
                        nodes_ids=matching_node_ids,
                        fields=uniqueness_constraint_fields,
                        typ=uniqueness_constraint_path.typ,
                    )
                )

        return violations

    async def check(self, node: Node, at: Timestamp | None = None, filters: list[str] | None = None) -> None:
        def _frozen_constraints(schema: MainSchemaTypes) -> frozenset[frozenset[str]]:
            if not schema.uniqueness_constraints:
                return frozenset()
            return frozenset(frozenset(uc) for uc in schema.uniqueness_constraints)

        node_schema = node.get_schema()
        include_node_schema = True
        frozen_node_constraints = _frozen_constraints(node_schema)
        schemas_to_check: list[MainSchemaTypes] = []
        if node_schema.inherit_from:
            for parent_schema_name in node_schema.inherit_from:
                parent_schema = self.schema_branch.get(name=parent_schema_name, duplicate=False)
                if not parent_schema.uniqueness_constraints:
                    continue
                schemas_to_check.append(parent_schema)
                frozen_parent_constraints = _frozen_constraints(parent_schema)
                if frozen_node_constraints <= frozen_parent_constraints:
                    include_node_schema = False

        if include_node_schema:
            schemas_to_check.append(node_schema)

        violations: list[UniquenessConstraintViolation] = []

        for schema in schemas_to_check:
            schema_filters = list(filters) if filters is not None else []
            for attr_schema in schema.attributes:
                if attr_schema.optional and attr_schema.unique and attr_schema.name not in schema_filters:
                    schema_filters.append(attr_schema.name)

            schema_violations = await self._get_single_schema_violations(
                node=node, node_schema=schema, at=at, filters=schema_filters
            )
            violations.extend(schema_violations)

        hfid_violations = [violation for violation in violations if violation.typ == UniquenessConstraintType.HFID]
        hfid_violation = hfid_violations[0] if len(hfid_violations) > 0 else None

        # If there are both a hfid violation and another one, in case of an upsert, we still want to update the node in case other violations are:
        # - either on subset fields of hfid, which would be necessarily violated too
        # - or on uniqueness constraints with a matching node id being the id of the hfid violation

        for violation in violations:
            if violation.typ == UniquenessConstraintType.HFID:
                continue

            if hfid_violation:
                if violation.typ == UniquenessConstraintType.SUBSET_OF_HFID:
                    continue

                if (
                    violation.typ == UniquenessConstraintType.STANDARD
                    and len(violation.nodes_ids) == 1
                    and next(iter(violation.nodes_ids)) == next(iter(hfid_violation.nodes_ids))
                ):
                    continue

            error_msg = f"Violates uniqueness constraint '{'-'.join(violation.fields)}'"
            raise ValidationError(error_msg)

        if hfid_violation:
            error_msg = f"Violates uniqueness constraint '{'-'.join(hfid_violation.fields)}'"
            raise HFIDViolatedError(error_msg, matching_nodes_ids=hfid_violation.nodes_ids)



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field


class RequestGeneratorRun(BaseModel):
    """Runs a generator."""

    generator_definition: GeneratorDefinitionModel = Field(..., description="The Generator definition")
    generator_instance: str | None = Field(
        default=None, description="The id of the generator instance if it previously existed"
    )
    commit: str = Field(..., description="The commit to target")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    repository_kind: str = Field(..., description="The kind of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    target_id: str = Field(..., description="The ID of the target object for this generator")
    target_name: str = Field(..., description="Name of the generator target")
    query: str = Field(..., description="The name of the query to use when collecting data")
    variables: dict = Field(..., description="Input variables when running the generator")


class RequestGeneratorDefinitionRun(BaseModel):
    """Sent to trigger a Generator to run on a specific branch."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    generator_definition: ProposedChangeGeneratorDefinition = Field(..., description="The Generator Definition")
    branch: str = Field(..., description="The branch to target")
    target_members: list[str] = Field(default_factory=list, description="List of targets to run the generator for")


class GeneratorDefinitionModel(BaseModel):
    definition_id: str = Field(..., description="The id of the generator definition.")
    definition_name: str = Field(..., description="The name of the generator definition.")
    query_name: str = Field(..., description="The name of the query to use when collecting data.")
    convert_query_response: bool = Field(
        ...,
        description="Decide if the generator should convert the result of the GraphQL query to SDK InfrahubNode objects.",
    )
    class_name: str = Field(..., description="The name of the generator class to run.")
    file_path: str = Field(..., description="The file path of the generator in the repository.")
    group_id: str = Field(..., description="The group to target when running this generator")
    parameters: dict = Field(..., description="The input parameters required to run this check")

    execute_in_proposed_change: bool = Field(
        ..., description="Indicates if the generator should execute in a proposed change."
    )
    execute_after_merge: bool = Field(..., description="Indicates if the generator should execute after a merge.")

    @classmethod
    def from_pc_generator_definition(cls, model: ProposedChangeGeneratorDefinition) -> GeneratorDefinitionModel:
        return GeneratorDefinitionModel(
            definition_id=model.definition_id,
            definition_name=model.definition_name,
            query_name=model.query_name,
            convert_query_response=model.convert_query_response,
            class_name=model.class_name,
            file_path=model.file_path,
            group_id=model.group_id,
            parameters=model.parameters,
            execute_in_proposed_change=model.execute_in_proposed_change,
            execute_after_merge=model.execute_after_merge,
        )


class ProposedChangeGeneratorDefinition(GeneratorDefinitionModel):
    query_models: list[str] = Field(..., description="The models to use when collecting data.")
    repository_id: str = Field(..., description="The id of the repository.")



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping

from infrahub import lock
from infrahub.core import registry
from infrahub.core.constants import RelationshipCardinality, RelationshipKind
from infrahub.core.constraint.node.runner import NodeConstraintRunner
from infrahub.core.node import Node
from infrahub.core.node.lock_utils import get_lock_names_on_object_mutation
from infrahub.core.protocols import CoreObjectTemplate
from infrahub.core.schema import GenericSchema
from infrahub.dependencies.registry import get_component_registry
from infrahub.lock import InfrahubMultiLock
from infrahub.profiles.node_applier import NodeProfilesApplier

if TYPE_CHECKING:
    from infrahub.core.branch import Branch
    from infrahub.core.relationship.model import RelationshipManager
    from infrahub.core.schema import MainSchemaTypes, NonGenericSchemaTypes, RelationshipSchema
    from infrahub.core.timestamp import Timestamp
    from infrahub.database import InfrahubDatabase


async def get_template_relationship_peers(
    db: InfrahubDatabase, template: CoreObjectTemplate, relationship: RelationshipSchema
) -> Mapping[str, CoreObjectTemplate]:
    """For a given relationship on the template, fetch the related peers."""
    template_relationship_manager: RelationshipManager = getattr(template, relationship.name)
    if relationship.cardinality == RelationshipCardinality.MANY:
        return await template_relationship_manager.get_peers(db=db, peer_type=CoreObjectTemplate)

    peers: dict[str, CoreObjectTemplate] = {}
    template_relationship_peer = await template_relationship_manager.get_peer(db=db, peer_type=CoreObjectTemplate)
    if template_relationship_peer:
        peers[template_relationship_peer.id] = template_relationship_peer
    return peers


async def extract_peer_data(
    db: InfrahubDatabase,
    template_peer: CoreObjectTemplate,
    obj_peer_schema: MainSchemaTypes,
    parent_obj: Node,
    current_template: CoreObjectTemplate,
) -> Mapping[str, Any]:
    obj_peer_data: dict[str, Any] = {}

    for attr_name in template_peer.get_schema().attribute_names:
        template_attr = getattr(template_peer, attr_name)
        if template_attr.value is None:
            continue
        if template_attr.is_default:
            # if template attr is_default and the value matches the object schema, then do not set the source
            try:
                if obj_peer_schema.get_attribute(name=attr_name).default_value == template_attr.value:
                    continue
            except ValueError:
                pass

        obj_peer_data[attr_name] = {"value": template_attr.value, "source": template_peer.id}

    for rel in template_peer.get_schema().relationship_names:
        rel_manager: RelationshipManager = getattr(template_peer, rel)

        if rel_manager.schema.name not in obj_peer_schema.relationship_names:
            continue

        peers_map = await rel_manager.get_peers(db=db)
        if rel_manager.schema.kind in [RelationshipKind.COMPONENT, RelationshipKind.PARENT] and list(
            peers_map.keys()
        ) == [current_template.id]:
            obj_peer_data[rel] = {"id": parent_obj.id}
            continue

        rel_peer_ids = []
        for peer_id, peer_object in peers_map.items():
            # deeper templates are handled in the next level of recursion
            if peer_object.get_schema().is_template_schema:
                continue
            rel_peer_ids.append({"id": peer_id})

        obj_peer_data[rel] = rel_peer_ids

    return obj_peer_data


async def handle_template_relationships(
    db: InfrahubDatabase,
    branch: Branch,
    obj: Node,
    template: CoreObjectTemplate,
    fields: list,
    constraint_runner: NodeConstraintRunner | None = None,
    at: Timestamp | None = None,
) -> None:
    if constraint_runner is None:
        component_registry = get_component_registry()
        constraint_runner = await component_registry.get_component(NodeConstraintRunner, db=db, branch=branch)

    for relationship in obj.get_relationships(kind=RelationshipKind.COMPONENT, exclude=fields):
        template_relationship_peers = await get_template_relationship_peers(
            db=db, template=template, relationship=relationship
        )
        if not template_relationship_peers:
            continue

        for template_relationship_peer in template_relationship_peers.values():
            # We retrieve peer schema for each peer in case we are processing a relationship which is based on a generic
            obj_peer_schema = registry.schema.get_node_schema(
                name=template_relationship_peer.get_schema().kind.removeprefix("Template"),
                branch=branch,
                duplicate=False,
            )
            obj_peer_data = await extract_peer_data(
                db=db,
                template_peer=template_relationship_peer,
                obj_peer_schema=obj_peer_schema,
                parent_obj=obj,
                current_template=template,
            )

            obj_peer = await Node.init(schema=obj_peer_schema, db=db, branch=branch, at=at)
            await obj_peer.new(db=db, **obj_peer_data)
            await constraint_runner.check(node=obj_peer, field_filters=list(obj_peer_data))
            await obj_peer.save(db=db)

            await handle_template_relationships(
                db=db,
                branch=branch,
                constraint_runner=constraint_runner,
                obj=obj_peer,
                template=template_relationship_peer,
                fields=fields,
                at=at,
            )


async def get_profile_ids(db: InfrahubDatabase, obj: Node) -> set[str]:
    if not hasattr(obj, "profiles"):
        return set()
    profile_rels = await obj.profiles.get_relationships(db=db)
    return {pr.peer_id for pr in profile_rels}


async def _do_create_node(
    node_class: type[Node],
    node_constraint_runner: NodeConstraintRunner,
    db: InfrahubDatabase,
    schema: NonGenericSchemaTypes,
    branch: Branch,
    fields_to_validate: list[str],
    data: dict[str, Any],
    at: Timestamp | None = None,
) -> Node:
    obj = await node_class.init(db=db, schema=schema, branch=branch)
    await obj.new(db=db, **data)
    await node_constraint_runner.check(node=obj, field_filters=fields_to_validate)
    await obj.save(db=db, at=at)

    object_template = await obj.get_object_template(db=db)
    if object_template:
        await handle_template_relationships(
            db=db,
            branch=branch,
            template=object_template,
            obj=obj,
            fields=fields_to_validate,
            at=at,
        )
    return obj


async def create_node(
    data: dict[str, Any],
    db: InfrahubDatabase,
    branch: Branch,
    schema: MainSchemaTypes,
    at: Timestamp | None = None,
) -> Node:
    """Create a node in the database if constraint checks succeed."""

    if isinstance(schema, GenericSchema):
        raise ValueError(f"Node of generic schema `{schema.name=}` can not be instantiated.")

    component_registry = get_component_registry()
    node_class = Node
    if schema.kind in registry.node:
        node_class = registry.node[schema.kind]

    fields_to_validate = list(data)

    preview_obj = await node_class.init(db=db, schema=schema, branch=branch)
    await preview_obj.new(db=db, process_pools=False, **data)
    schema_branch = db.schema.get_schema_branch(name=branch.name)
    lock_names = get_lock_names_on_object_mutation(node=preview_obj, schema_branch=schema_branch)

    obj: Node
    async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names, metrics=False):
        if db.is_transaction:
            node_constraint_runner = await component_registry.get_component(NodeConstraintRunner, db=db, branch=branch)

            obj = await _do_create_node(
                node_class=node_class,
                node_constraint_runner=node_constraint_runner,
                db=db,
                schema=schema,
                branch=branch,
                fields_to_validate=fields_to_validate,
                data=data,
                at=at,
            )
        else:
            async with db.start_transaction() as dbt:
                node_constraint_runner = await component_registry.get_component(
                    NodeConstraintRunner, db=dbt, branch=branch
                )

                obj = await _do_create_node(
                    node_class=node_class,
                    node_constraint_runner=node_constraint_runner,
                    db=dbt,
                    schema=schema,
                    branch=branch,
                    fields_to_validate=fields_to_validate,
                    data=data,
                    at=at,
                )

    if await get_profile_ids(db=db, obj=obj):
        node_profiles_applier = NodeProfilesApplier(db=db, branch=branch)
        await node_profiles_applier.apply_profiles(node=obj)
        await obj.save(db=db)

    return obj



from __future__ import annotations

from .attribute import (
    AnyAttributeType,
    AttributeInterface,
    BaseAttribute,
    BoolAttributeType,
    CheckboxAttributeType,
    DropdownFields,
    DropdownType,
    IntAttributeType,
    IPHostType,
    IPNetworkType,
    JSONAttributeType,
    ListAttributeType,
    MacAddressType,
    NumberAttributeType,
    RelatedIPAddressNodeInput,
    RelatedIPPrefixNodeInput,
    RelatedNodeInput,
    StrAttributeType,
    TextAttributeType,
)
from .branch import BranchType
from .interface import InfrahubInterface
from .node import InfrahubObject
from .permission import PaginatedObjectPermission
from .relationship import RelationshipNode
from .standard_node import InfrahubObjectType
from .task import TaskNodes
from .task_log import TaskLog, TaskLogEdge, TaskLogNodes

__all__ = [
    "AnyAttributeType",
    "AttributeInterface",
    "BaseAttribute",
    "BoolAttributeType",
    "BranchType",
    "CheckboxAttributeType",
    "DropdownFields",
    "DropdownType",
    "IPHostType",
    "IPNetworkType",
    "InfrahubInterface",
    "InfrahubObject",
    "InfrahubObjectType",
    "IntAttributeType",
    "JSONAttributeType",
    "ListAttributeType",
    "MacAddressType",
    "NumberAttributeType",
    "PaginatedObjectPermission",
    "RelatedIPAddressNodeInput",
    "RelatedIPPrefixNodeInput",
    "RelatedNodeInput",
    "RelationshipNode",
    "StrAttributeType",
    "TaskLog",
    "TaskLogEdge",
    "TaskLogNodes",
    "TaskNodes",
    "TextAttributeType",
]


RELATIONS_PROPERTY_MAP: dict[str, str] = {
    "is_visible": "_relation__is_visible",
    "is_protected": "_relation__is_protected",
    "owner": "_relation__owner",
    "source": "_relation__source",
    "updated_at": "_relation__updated_at",
    "__typename": "__typename",
}

RELATIONS_PROPERTY_MAP_REVERSED = {value: key for key, value in RELATIONS_PROPERTY_MAP.items()}



from typing import TYPE_CHECKING

from infrahub.core import registry
from infrahub.core.branch import Branch
from infrahub.core.node import Node
from infrahub.core.timestamp import Timestamp
from infrahub.database import InfrahubDatabase
from infrahub.exceptions import ValidationError

from .interface import NodeConstraintInterface

if TYPE_CHECKING:
    from infrahub.core.schema import MainSchemaTypes


class NodeAttributeUniquenessConstraint(NodeConstraintInterface):
    def __init__(self, db: InfrahubDatabase, branch: Branch) -> None:
        self.db = db
        self.branch = branch

    async def check(self, node: Node, at: Timestamp | None = None, filters: list[str] | None = None) -> None:
        at = Timestamp(at)
        node_schema = node.get_schema()
        for unique_attr in node_schema.unique_attributes:
            if filters and unique_attr.name not in filters:
                continue

            comparison_schema: MainSchemaTypes = node_schema
            attr = getattr(node, unique_attr.name)
            if unique_attr.inherited:
                for generic_parent_schema_name in node_schema.inherit_from:
                    generic_parent_schema = self.db.schema.get(
                        generic_parent_schema_name, branch=self.branch, duplicate=False
                    )
                    parent_attr = generic_parent_schema.get_attribute_or_none(unique_attr.name)
                    if parent_attr is None:
                        continue
                    if parent_attr.unique is True:
                        comparison_schema = generic_parent_schema
                        break
            nodes = await registry.manager.query(
                schema=comparison_schema,
                filters={f"{unique_attr.name}__value": attr.value},
                fields={},
                db=self.db,
                branch=self.branch,
                at=at,
            )

            if any(n for n in nodes if n.get_id() != node.id):
                raise ValidationError(
                    {unique_attr.name: f"An object already exist with this value: {unique_attr.name}: {attr.value}"}
                )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

from graphene import InputObjectType, Mutation
from graphene.types.mutation import MutationOptions
from infrahub_sdk.utils import extract_fields_first_node
from typing_extensions import Self

from infrahub import config, lock
from infrahub.core.constants import MutationAction
from infrahub.core.constraint.node.runner import NodeConstraintRunner
from infrahub.core.manager import NodeManager
from infrahub.core.node.create import create_node, get_profile_ids
from infrahub.core.schema import MainSchemaTypes, NodeSchema
from infrahub.core.schema.generic_schema import GenericSchema
from infrahub.core.schema.profile_schema import ProfileSchema
from infrahub.core.schema.template_schema import TemplateSchema
from infrahub.core.timestamp import Timestamp
from infrahub.database import retry_db_transaction
from infrahub.dependencies.registry import get_component_registry
from infrahub.events.generator import generate_node_mutation_events
from infrahub.exceptions import HFIDViolatedError, InitializationError, NodeNotFoundError
from infrahub.graphql.context import apply_external_context
from infrahub.graphql.field_extractor import extract_graphql_fields
from infrahub.lock import InfrahubMultiLock
from infrahub.log import get_log_data, get_logger
from infrahub.profiles.node_applier import NodeProfilesApplier

from ...core.node.lock_utils import get_lock_names_on_object_mutation
from .node_getter.by_default_filter import MutationNodeGetterByDefaultFilter

if TYPE_CHECKING:
    from graphql import GraphQLResolveInfo

    from infrahub.core.branch import Branch
    from infrahub.core.node import Node
    from infrahub.database import InfrahubDatabase
    from infrahub.graphql.types.context import ContextInput

    from ..initialization import GraphqlContext


log = get_logger()


@dataclass
class DeleteResult:
    node: Node
    mutation: InfrahubMutationMixin
    deleted_nodes: list[Node] = field(default_factory=list)


# ------------------------------------------
# Infrahub GraphQLType
# ------------------------------------------
class InfrahubMutationOptions(MutationOptions):
    schema: MainSchemaTypes | None = None

    @property
    def active_schema(self) -> MainSchemaTypes:
        if self.schema:
            return self.schema
        raise InitializationError("This class is not initialized with a schema")


class InfrahubMutationMixin:
    _meta: InfrahubMutationOptions

    @classmethod
    async def mutate(
        cls,
        root: dict,  # noqa: ARG003
        info: GraphQLResolveInfo,
        data: InputObjectType,
        context: ContextInput | None = None,
        **kwargs: dict[str, Any],
    ) -> Self:
        graphql_context: GraphqlContext = info.context
        await apply_external_context(graphql_context=graphql_context, context_input=context)

        obj = None
        mutation = None
        action = MutationAction.UNDEFINED
        deleted_nodes: list[Node] = []

        if "Create" in cls.__name__:
            obj, mutation = await cls.mutate_create(info=info, branch=graphql_context.branch, data=data)
            action = MutationAction.CREATED
        elif "Update" in cls.__name__:
            obj, mutation = await cls.mutate_update(info=info, branch=graphql_context.branch, data=data, **kwargs)
            action = MutationAction.UPDATED
        elif "Upsert" in cls.__name__:
            node_manager = NodeManager()
            node_getter_default_filter = MutationNodeGetterByDefaultFilter(
                db=graphql_context.db, node_manager=node_manager
            )
            obj, mutation, created = await cls.mutate_upsert(
                info=info,
                branch=graphql_context.branch,
                data=data,
                node_getter_default_filter=node_getter_default_filter,
                **kwargs,
            )
            if created:
                action = MutationAction.CREATED
            else:
                action = MutationAction.UPDATED
        elif "Delete" in cls.__name__:
            delete_result = await cls.mutate_delete(info=info, branch=graphql_context.branch, data=data, **kwargs)
            obj = delete_result.node
            mutation = delete_result.mutation
            deleted_nodes = delete_result.deleted_nodes

            action = MutationAction.DELETED
        else:
            raise ValueError(
                f"Unexpected class Name: {cls.__name__}, should end with Create, Update, Upsert, or Delete"
            )

        # Reset the time of the query to guarantee that all resolvers executed after this point will account for the changes
        graphql_context.at = Timestamp()

        if config.SETTINGS.broker.enable and graphql_context.background and obj.node_changelog.has_changes:
            log_data = get_log_data()
            request_id = log_data.get("request_id", "")

            events = await generate_node_mutation_events(
                node=obj,
                deleted_nodes=deleted_nodes,
                db=graphql_context.db,
                branch=graphql_context.branch,
                context=graphql_context.get_context(),
                request_id=request_id,
                action=action,
            )

            for event in events:
                graphql_context.background.add_task(graphql_context.active_service.event.send, event)

        return mutation

    @classmethod
    async def mutate_create(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        override_data: dict[str, Any] | None = None,
    ) -> tuple[Node, Self]:
        db = database or info.context.db
        schema = cls._meta.active_schema

        create_data = dict(data)
        create_data.update(override_data or {})

        obj = await create_node(
            data=create_data,
            db=db,
            branch=branch,
            schema=schema,
        )

        graphql_response = await build_graphql_response(info=info, db=db, obj=obj)
        return obj, cls(**graphql_response)

    @classmethod
    async def _call_mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        db: InfrahubDatabase,
        obj: Node,
        skip_uniqueness_check: bool = False,
    ) -> tuple[Node, Self]:
        """
        Wrapper around mutate_update to potentially activate locking and call it within a database transaction.
        """

        # Prepare a clone to compute locks without triggering pool allocations
        preview_obj = await NodeManager.get_one_by_id_or_default_filter(
            db=db,
            kind=obj.get_kind(),
            id=obj.get_id(),
            branch=branch,
        )
        await preview_obj.from_graphql(db=db, data=data, process_pools=False)

        schema_branch = db.schema.get_schema_branch(name=branch.name)
        lock_names = get_lock_names_on_object_mutation(node=preview_obj, schema_branch=schema_branch)

        # FIXME: do not lock when data does not contain uniqueness constraint fields or resource pool allocations
        async with InfrahubMultiLock(lock_registry=lock.registry, locks=lock_names, metrics=False):
            if db.is_transaction:
                obj = await cls.mutate_update_object(
                    db=db, info=info, data=data, branch=branch, obj=obj, skip_uniqueness_check=skip_uniqueness_check
                )

                result = await cls.mutate_update_to_graphql(db=db, info=info, obj=obj)
                return obj, result

            async with db.start_transaction() as dbt:
                obj = await cls.mutate_update_object(
                    db=dbt,
                    info=info,
                    data=data,
                    branch=branch,
                    obj=obj,
                    skip_uniqueness_check=skip_uniqueness_check,
                )

                result = await cls.mutate_update_to_graphql(db=dbt, info=info, obj=obj)
                return obj, result

    @classmethod
    @retry_db_transaction(name="object_update")
    async def mutate_update(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        database: InfrahubDatabase | None = None,
        node: Node | None = None,
    ) -> tuple[Node, Self]:
        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db

        obj = node or await NodeManager.find_object(
            db=db, kind=cls._meta.active_schema.kind, id=data.get("id"), hfid=data.get("hfid"), branch=branch
        )
        obj, result = await cls._call_mutate_update(info=info, data=data, db=db, branch=branch, obj=obj)

        return obj, result

    @classmethod
    async def mutate_update_object(
        cls,
        db: InfrahubDatabase,
        info: GraphQLResolveInfo,  # noqa: ARG003
        data: InputObjectType,
        branch: Branch,
        obj: Node,
        skip_uniqueness_check: bool = False,
    ) -> Node:
        component_registry = get_component_registry()
        node_constraint_runner = await component_registry.get_component(NodeConstraintRunner, db=db, branch=branch)

        await obj.from_graphql(db=db, data=data)
        fields_to_validate = list(data)
        await node_constraint_runner.check(
            node=obj, field_filters=fields_to_validate, skip_uniqueness_check=skip_uniqueness_check
        )

        fields = list(data.keys())
        for field_to_remove in ("id", "hfid"):
            if field_to_remove in fields:
                fields.remove(field_to_remove)

        after_mutate_profile_ids = await get_profile_ids(db=db, obj=obj)
        if after_mutate_profile_ids or (not after_mutate_profile_ids and obj.uses_profiles()):
            node_profiles_applier = NodeProfilesApplier(db=db, branch=branch)
            updated_field_names = await node_profiles_applier.apply_profiles(node=obj)
            fields += updated_field_names
        await obj.save(db=db, fields=fields)

        return obj

    @classmethod
    async def mutate_update_to_graphql(
        cls,
        db: InfrahubDatabase,
        info: GraphQLResolveInfo,
        obj: Node,
    ) -> Self:
        fields_object = extract_graphql_fields(info=info)
        fields_object = fields_object.get("object", {})
        result: dict[str, Any] = {"ok": True}
        if fields_object:
            result["object"] = await obj.to_graphql(db=db, fields=fields_object)
        return cls(**result)

    @classmethod
    @retry_db_transaction(name="object_upsert")
    async def mutate_upsert(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
        node_getter_default_filter: MutationNodeGetterByDefaultFilter,
        database: InfrahubDatabase | None = None,
    ) -> tuple[Node, Self, bool]:
        """
        First, check whether payload contains data identifying the node, such as id, hfid, or relevant fields for
        default_filter. If not, we will try to create the node, but this creation might fail if payload contains
        hfid fields (not `hfid` field itself) that would match an existing node in the database. In that case,
        we would update the node without rerunning uniqueness constraint.
        """

        schema = cls._meta.active_schema
        schema_name = schema.kind

        graphql_context: GraphqlContext = info.context
        db = database or graphql_context.db
        dict_data = dict(data)
        node = None

        if "id" in dict_data:
            node = await NodeManager.get_one(
                db=db, id=dict_data["id"], kind=schema_name, branch=branch, raise_on_error=True
            )
            updated_obj, mutation = await cls._call_mutate_update(
                info=info,
                data=data,
                db=db,
                branch=branch,
                obj=node,
            )
            return updated_obj, mutation, False

        if not schema.human_friendly_id and schema.default_filter is not None:
            node = await node_getter_default_filter.get_node(node_schema=schema, data=data, branch=branch)

        if "hfid" in data:
            node = await NodeManager.get_one_by_hfid(db=db, hfid=dict_data["hfid"], kind=schema_name, branch=branch)

        if node is not None:
            updated_obj, mutation = await cls._call_mutate_update(
                info=info,
                data=data,
                db=db,
                branch=branch,
                obj=node,
            )
            return updated_obj, mutation, False

        try:
            # This is a hack to avoid sitatuions where a node has an attribute or relationship called "pop"
            # which would have overridden the `pop` method of the InputObjectType object and as such would have
            # caused an error when trying to call `data.pop("hfid", None)`.
            # TypeError: 'NoneType' object is not callable
            data._pop = dict.pop.__get__(data, dict)
            data._pop("hfid", None)  # `hfid` is invalid for creation.
            created_obj, mutation = await cls.mutate_create(info=info, data=data, branch=branch)
            return created_obj, mutation, True
        except HFIDViolatedError as exc:
            # Only the HFID constraint has been violated, it means the node exists and we can update without rerunning constraints
            if len(exc.matching_nodes_ids) > 1:
                raise RuntimeError(f"Multiple {schema_name} nodes have the same hfid") from exc
            node_id = list(exc.matching_nodes_ids)[0]

            try:
                node = await NodeManager.get_one(
                    db=db, id=node_id, kind=schema_name, branch=branch, raise_on_error=True
                )
            except NodeNotFoundError as exc:
                if branch.is_default:
                    raise
                raise NodeNotFoundError(
                    node_type=exc.node_type,
                    identifier=exc.identifier,
                    branch_name=branch.name,
                    message=(
                        f"Node {exc.identifier} / {exc.node_type} uses this human-friendly ID, but does not exist on"
                        f" this branch. Please rebase this branch to access {exc.identifier} / {exc.node_type}"
                    ),
                ) from exc
            updated_obj, mutation = await cls._call_mutate_update(
                info=info,
                data=data,
                db=db,
                branch=branch,
                obj=node,
                skip_uniqueness_check=True,
            )
            return updated_obj, mutation, False

    @classmethod
    async def _delete_obj(cls, graphql_context: GraphqlContext, branch: Branch, obj: Node) -> list[Node]:
        db = graphql_context.db
        async with db.start_transaction() as dbt:
            deleted = await NodeManager.delete(db=dbt, branch=branch, nodes=[obj])
        deleted_str = ", ".join([f"{d.get_kind()}({d.get_id()})" for d in deleted])
        log.info(f"nodes deleted: {deleted_str}")
        return deleted

    @classmethod
    @retry_db_transaction(name="object_delete")
    async def mutate_delete(
        cls,
        info: GraphQLResolveInfo,
        data: InputObjectType,
        branch: Branch,
    ) -> DeleteResult:
        graphql_context: GraphqlContext = info.context

        obj = await NodeManager.find_object(
            db=graphql_context.db,
            kind=cls._meta.active_schema.kind,
            id=data.get("id"),
            hfid=data.get("hfid"),
            branch=branch,
        )

        deleted = await cls._delete_obj(graphql_context=graphql_context, branch=branch, obj=obj)

        ok = True

        return DeleteResult(node=obj, mutation=cls(ok=ok), deleted_nodes=deleted)


class InfrahubMutation(InfrahubMutationMixin, Mutation):
    @classmethod
    def __init_subclass_with_meta__(
        cls,
        schema: NodeSchema | GenericSchema | ProfileSchema | TemplateSchema | None = None,
        _meta: InfrahubMutationOptions | None = None,
        **options: dict[str, Any],
    ) -> None:
        # Make sure schema is a valid NodeSchema Node Class
        if not isinstance(schema, NodeSchema | GenericSchema | ProfileSchema | TemplateSchema):
            raise ValueError(f"You need to pass a valid NodeSchema in '{cls.__name__}.Meta', received '{schema}'")

        if not _meta:
            _meta = InfrahubMutationOptions(cls)

        _meta.schema = schema

        super().__init_subclass_with_meta__(_meta=_meta, **options)


def _get_data_fields(data: InputObjectType) -> list[str]:
    return [field for field in data.keys() if field not in ["id", "hfid"]]


async def build_graphql_response(info: GraphQLResolveInfo, db: InfrahubDatabase, obj: Node) -> dict:
    fields = await extract_fields_first_node(info)
    result: dict[str, Any] = {"ok": True}
    if "object" in fields:
        result["object"] = await obj.to_graphql(db=db, fields=fields.get("object", {}))
    return result



import logging
from typing import TYPE_CHECKING

import typer
import uvicorn

if TYPE_CHECKING:
    from infrahub.cli.context import CliContext

app = typer.Typer()


@app.callback()
def callback() -> None:
    """
    Control the API Server.
    """


log_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "format": "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        },
    },
    "handlers": {
        "default": {"formatter": "default", "class": "logging.NullHandler"},
        "access": {"formatter": "access", "class": "logging.NullHandler"},
    },
    "loggers": {
        "uvicorn.error": {"level": "INFO", "handlers": ["default"], "propagate": True},
        "uvicorn.access": {"level": "INFO", "handlers": ["access"], "propagate": False},
    },
}


@app.command()
def start(
    ctx: typer.Context,
    listen: str = typer.Option("127.0.0.1", help="Address used to listen for new request."),
    port: int = typer.Option(8000, help="Port used to listen for new request."),
    debug: bool = typer.Option(False, help="Enable advanced logging and troubleshooting"),
) -> None:
    """Start Infrahub in Debug Mode with reload enabled."""

    # it's not possible to pass the location of the config file directly to uvicorn.run
    # so we must rely on the environment variable

    logging.getLogger("neo4j").setLevel(logging.ERROR)

    context: CliContext = ctx.obj

    if debug:
        uvicorn.run(
            context.application,
            host=listen,
            port=port,
            log_level="info",
            reload=True,
            reload_excludes=["examples", "repositories"],
            log_config=log_config,
        )
    else:
        uvicorn.run(
            context.application,
            host=listen,
            port=port,
            log_level="info",
            log_config=log_config,
        )


# gunicorn infrahub.server:app --workers 4 --worker-class uvicorn.workers.UvicornWorker --bind 0.0.0.0:8000



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import ujson

from infrahub.core.query import Query, QueryType
from infrahub.core.timestamp import Timestamp

if TYPE_CHECKING:
    from infrahub.database import InfrahubDatabase


class SchemaAttributeUpdateQuery(Query):
    name = "schema_attribute_update"
    type = QueryType.WRITE
    insert_return = True

    def __init__(
        self,
        attribute_name: str,
        node_name: str,
        node_namespace: str,
        new_value: Any,
        previous_value: Any | None = None,
        **kwargs: Any,
    ):
        self.attr_name = attribute_name
        self.node_name = node_name
        self.node_namespace = node_namespace
        self.attr_new_value = new_value
        self.attr_previous_value = previous_value

        super().__init__(**kwargs)

    def render_match(self) -> str:
        return self._render_match_schema_node()

    def _render_match_schema_node(self) -> str:
        # ruff: noqa: E501
        query = """
        MATCH path = (node:SchemaNode)-[:HAS_ATTRIBUTE]->(attr:Attribute)-[rel:HAS_VALUE]->(av:AttributeValue)
        """
        return query

    def _render_match_schema_attribute(self) -> str:
        # ruff: noqa: E501
        query = """
        MATCH path = (node:SchemaNode)-[:IS_RELATED]->(:Relationship)<-[:IS_RELATED]-(attr_node:SchemaAttribute)-[:HAS_ATTRIBUTE]->(attr:Attribute)-[rel:HAS_VALUE]->(av:AttributeValue)
        """
        return query

    def render_where(self) -> str:
        at = self.at or Timestamp()
        filters, params = at.get_query_filter_path()
        self.params.update(params)

        # ruff: noqa: E501
        query = """
        WHERE exists((node)-[:HAS_ATTRIBUTE]->(:Attribute {name: "name"})-[:HAS_VALUE]->(:AttributeValue { value: $node_name}))
            AND exists((node)-[:HAS_ATTRIBUTE]->(:Attribute {name: "namespace"})-[:HAS_VALUE]->(:AttributeValue { value: $node_namespace }))
            AND attr.name = $attr_name
            AND all(r IN relationships(path) WHERE ( %(filters)s ))
        """ % {"filters": filters}
        if self.attr_previous_value:
            query += "\n  AND av.value = $attr_previous_value"

        return query

    async def query_init(self, db: InfrahubDatabase, **kwargs: dict[str, Any]) -> None:  # noqa: ARG002
        self.params["attr_name"] = self.attr_name
        self.params["attr_new_value"] = (
            ujson.dumps(self.attr_new_value) if isinstance(self.attr_new_value, list) else self.attr_new_value
        )
        self.params["attr_previous_value"] = (
            ujson.dumps(self.attr_previous_value)
            if isinstance(self.attr_previous_value, list)
            else self.attr_previous_value
        )
        self.params["node_name"] = self.node_name
        self.params["node_namespace"] = self.node_namespace

        # Select the Node to update, either SchemaNode or SchemaAttribute
        self.add_to_query(self.render_match())
        self.add_to_query(self.render_where())

        query = """
        MERGE (new_value: AttributeValue { value: $attr_new_value, is_default: false })
        CREATE (attr)-[:HAS_VALUE { branch: rel.branch, branch_level: rel.branch_level, status: "active", from: $at } ]->(new_value)
        SET rel.to = $at
        """
        self.add_to_query(query)
        self.return_labels = ["attr"]



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import typer
from deepdiff import DeepDiff
from infrahub_sdk.async_typer import AsyncTyper
from prefect.client.orchestration import get_client

from infrahub import config
from infrahub.core.initialization import (
    create_anonymous_role,
    create_default_account_groups,
    get_root_node,
    initialize_registry,
)
from infrahub.core.manager import NodeManager
from infrahub.core.migrations.shared import get_migration_console
from infrahub.core.protocols import CoreAccount, CoreObjectPermission
from infrahub.dependencies.registry import build_component_registry
from infrahub.lock import initialize_lock
from infrahub.menu.menu import default_menu
from infrahub.menu.models import MenuDict
from infrahub.menu.repository import MenuRepository
from infrahub.menu.utils import create_default_menu
from infrahub.trigger.tasks import trigger_configure_all
from infrahub.workflows.initialization import (
    setup_blocks,
    setup_deployments,
    setup_worker_pools,
)

from .db import (
    detect_migration_to_run,
    initialize_internal_schema,
    mark_branches_needing_rebase,
    migrate_database,
    trigger_rebase_branches,
    update_core_schema,
)

if TYPE_CHECKING:
    from infrahub.cli.context import CliContext
    from infrahub.core.branch.models import Branch
    from infrahub.database import InfrahubDatabase

app = AsyncTyper()
console = get_migration_console()


@app.command(name="upgrade")
async def upgrade_cmd(
    ctx: typer.Context,
    config_file: str = typer.Argument("infrahub.toml", envvar="INFRAHUB_CONFIG"),
    check: bool = typer.Option(False, help="Check the state of the system without upgrading."),
    rebase_branches: bool = typer.Option(False, help="Rebase and apply migrations to branches if required."),
    interactive: bool = typer.Option(
        False, help="Use interactive prompt to accept or deny rebase of individual branches."
    ),
) -> None:
    """Upgrade Infrahub to the latest version."""

    logging.getLogger("infrahub").setLevel(logging.WARNING)
    logging.getLogger("neo4j").setLevel(logging.ERROR)
    logging.getLogger("prefect").setLevel(logging.ERROR)
    os.environ["PREFECT_SERVER_ANALYTICS_ENABLED"] = "false"

    config.load_and_exit(config_file_name=config_file)

    context: CliContext = ctx.obj
    dbdriver = await context.init_db(retry=1)

    await initialize_registry(db=dbdriver)
    initialize_lock()

    build_component_registry()

    root_node = await get_root_node(db=dbdriver)

    # NOTE add step to validate if the database and the task manager are reachable

    # -------------------------------------------
    # Add pre-upgrade  validation
    # -------------------------------------------

    # -------------------------------------------
    # Upgrade Infrahub Database and Schema
    # -------------------------------------------

    migrations = await detect_migration_to_run(current_graph_version=root_node.graph_version)
    if check:
        await dbdriver.close()
        return

    if not await migrate_database(db=dbdriver, initialize=False, migrations=migrations):
        # A migration failed, stop the upgrade process
        console.log("Upgrade cancelled due to migration failure.")
        await dbdriver.close()
        return

    await initialize_internal_schema()
    await update_core_schema(db=dbdriver, initialize=False)

    # -------------------------------------------
    # Upgrade Internal Objects, generated and managed by Infrahub
    # -------------------------------------------
    await upgrade_menu(db=dbdriver)
    await upgrade_permissions(db=dbdriver)

    # -------------------------------------------
    # Upgrade External system : Task Manager
    # -------------------------------------------
    async with get_client(sync_client=False) as client:
        await setup_blocks()
        await setup_worker_pools(client=client)
        await setup_deployments(client=client)
        await trigger_configure_all()

    # -------------------------------------------
    # Perform branch rebase and apply migrations to them
    # -------------------------------------------
    branches = await mark_branches_needing_rebase(db=dbdriver)
    plural = len(branches) != 1
    get_migration_console().log(
        f"Found {len(branches)} {'branches' if plural else 'branch'} that {'need' if plural else 'needs'} to be rebased"
    )

    if rebase_branches:
        branches_to_rebase: list[Branch] = []
        if not interactive:
            branches_to_rebase = branches
        else:
            for branch in branches:
                if typer.confirm(f"Rebase branch {branch.name}?"):
                    branches_to_rebase.append(branch)

        await trigger_rebase_branches(db=dbdriver, branches=branches_to_rebase)

    await dbdriver.close()


async def upgrade_menu(db: InfrahubDatabase) -> None:
    menu_repository = MenuRepository(db=db)
    menu_nodes = await menu_repository.get_menu_db()
    menu_items = await menu_repository.get_menu(nodes=menu_nodes)
    default_menu_dict = MenuDict.from_definition_list(default_menu)

    if not menu_nodes:
        await create_default_menu(db=db)
        return

    diff_menu = DeepDiff(menu_items.to_rest(), default_menu_dict.to_rest(), ignore_order=True)

    if not diff_menu:
        console.log("Menu Up to date, nothing to update")
        return

    await menu_repository.update_menu(existing_menu=menu_items, new_menu=default_menu_dict, menu_nodes=menu_nodes)
    console.log("Menu has been updated")


async def upgrade_permissions(db: InfrahubDatabase) -> None:
    existing_permissions = await NodeManager.query(schema=CoreObjectPermission, db=db, limit=1)
    if existing_permissions:
        console.log("Permissions Up to date, nothing to update")
        return

    await setup_permissions(db=db)
    console.log("Permissions have been updated")


async def setup_permissions(db: InfrahubDatabase) -> None:
    existing_accounts = await NodeManager.query(schema=CoreAccount, db=db, limit=1)
    await create_default_account_groups(db=db, admin_accounts=existing_accounts)

    if config.SETTINGS.main.allow_anonymous_access:
        await create_anonymous_role(db=db)



from infrahub.exceptions import Error


class QueryValidationError(Error):
    HTTP_CODE = 400

    def __init__(self, message: str):
        self.message = message


class SchemaNotValidError(Error):
    HTTP_CODE = 422

    def __init__(self, message: str):
        self.message = message



from infrahub.core.branch import Branch
from infrahub.core.branch.enums import BranchStatus


def raise_needs_rebase_error(branch_name: str) -> None:
    raise ValueError(f"Branch {branch_name} must be rebased before any updates can be made")


def check_need_rebase_status(branch: Branch) -> None:
    if branch.status == BranchStatus.NEED_REBASE:
        raise_needs_rebase_error(branch_name=branch.name)



from pydantic import BaseModel, ConfigDict, Field

from infrahub.context import InfrahubContext
from infrahub.message_bus.types import ProposedChangeBranchDiff


class RequestArtifactDefinitionGenerate(BaseModel):
    """Sent to trigger the generation of artifacts for a given branch."""

    artifact_definition_id: str = Field(..., description="The unique ID of the Artifact Definition")
    artifact_definition_name: str = Field(..., description="The name of the Artifact Definition")
    branch: str = Field(..., description="The branch to target")
    limit: list[str] = Field(
        default_factory=list,
        description="List of targets to limit the scope of the generation, if populated only the included artifacts will be regenerated",
    )


class RequestArtifactGenerate(BaseModel):
    """Runs to generate an artifact"""

    artifact_name: str = Field(..., description="Name of the artifact")
    artifact_definition: str = Field(..., description="The ID of the artifact definition")
    artifact_definition_name: str = Field(..., description="The name of the artifact definition")
    commit: str = Field(..., description="The commit to target")
    content_type: str = Field(..., description="Content type of the artifact")
    transform_type: str = Field(..., description="The type of transform associated with this artifact")
    transform_location: str = Field(..., description="The transforms location within the repository")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    repository_kind: str = Field(..., description="The kind of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    convert_query_response: bool = Field(
        default=False,
        description="Indicate if the query response should be converted to InfrahubNode objects for Python transforms",
    )
    target_id: str = Field(..., description="The ID of the target object for this artifact")
    target_kind: str = Field(..., description="The kind of the target object for this artifact")
    target_name: str = Field(..., description="Name of the artifact target")
    artifact_id: str | None = Field(default=None, description="The id of the artifact if it previously existed")
    query: str = Field(..., description="The name of the query to use when collecting data")  # Deprecated
    query_id: str = Field(..., description="The id of the query to use when collecting data")
    timeout: int = Field(..., description="Timeout for requests used to generate this artifact")
    variables: dict = Field(..., description="Input variables when generating the artifact")
    context: InfrahubContext = Field(..., description="The context of the task")


class GitRepositoryAdd(BaseModel):
    """Clone and sync an external repository after creation."""

    location: str = Field(..., description="The external URL of the repository")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    created_by: str | None = Field(default=None, description="The user ID of the user that created the repository")
    default_branch_name: str | None = Field(None, description="Default branch for this repository")
    infrahub_branch_name: str = Field(..., description="Infrahub branch on which to sync the remote repository")
    infrahub_branch_id: str = Field(..., description="Id of the Infrahub branch on which to sync the remote repository")
    internal_status: str = Field(..., description="Administrative status of the repository")


class GitRepositoryAddReadOnly(BaseModel):
    """Clone and sync an external repository after creation."""

    location: str = Field(..., description="The external URL of the repository")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    ref: str = Field(..., description="Ref to track on the external repository")
    created_by: str | None = Field(default=None, description="The user ID of the user that created the repository")
    infrahub_branch_name: str = Field(..., description="Infrahub branch on which to sync the remote repository")
    infrahub_branch_id: str = Field(..., description="Id of the Infrahub branch on which to sync the remote repository")
    internal_status: str = Field(..., description="Internal status of the repository")


class GitRepositoryPullReadOnly(BaseModel):
    """Update a read-only repository to the latest commit for its ref"""

    location: str = Field(..., description="The external URL of the repository")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    ref: str | None = Field(None, description="Ref to track on the external repository")
    commit: str | None = Field(None, description="Specific commit to pull")
    infrahub_branch_name: str = Field(..., description="Infrahub branch on which to sync the remote repository")
    infrahub_branch_id: str = Field(..., description="Infrahub branch on which to sync the remote repository")


class GitRepositoryMerge(BaseModel):
    """Merge one branch into another."""

    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    internal_status: str = Field(..., description="Administrative status of the repository")
    source_branch: str = Field(..., description="The source branch")
    destination_branch: str = Field(..., description="The destination branch")
    destination_branch_id: str = Field(..., description="The ID of the destination branch")
    default_branch: str = Field(..., description="The default branch in Git")


class GitRepositoryImportObjects(BaseModel):
    """Re run import job against an existing commit."""

    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    repository_kind: str = Field(..., description="The type of repository")
    commit: str = Field(..., description="Specific commit to pull")
    infrahub_branch_name: str = Field(..., description="Infrahub branch on which to sync the remote repository")


class GitDiffNamesOnly(BaseModel):
    """Request a list of modified files between two commits."""

    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the repository")
    repository_kind: str = Field(..., description="The kind of the repository")
    first_commit: str = Field(..., description="The first commit")
    second_commit: str | None = Field(None, description="The second commit")


class GitDiffNamesOnlyResponse(BaseModel):
    files_added: list[str] = Field(..., description="Files added")
    files_changed: list[str] = Field(..., description="Files changed")
    files_removed: list[str] = Field(..., description="Files removed")


class UserCheckDefinitionData(BaseModel):
    """Triggers user defined checks to run based on a Check Definition."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    check_definition_id: str = Field(..., description="The unique ID of the check definition")
    commit: str = Field(..., description="The commit to target")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    file_path: str = Field(..., description="The path and filename of the check")
    class_name: str = Field(..., description="The name of the class containing the check")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    branch_diff: ProposedChangeBranchDiff = Field(..., description="The calculated diff between the two branches")


class UserCheckData(BaseModel):
    """Runs a check as defined within a CoreCheckDefinition within a repository."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    validator_id: str = Field(..., description="The id of the validator associated with this check")
    validator_execution_id: str = Field(..., description="The id of current execution of the associated validator")
    check_execution_id: str = Field(..., description="The unique ID for the current execution of this check")
    check_definition_id: str = Field(..., description="The unique ID of the check definition")
    commit: str = Field(..., description="The commit to target")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    branch_name: str = Field(..., description="The branch where the check is run")
    file_path: str = Field(..., description="The path and filename of the check")
    class_name: str = Field(..., description="The name of the class containing the check")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    variables: dict = Field(default_factory=dict, description="Input variables when running the check")
    name: str = Field(..., description="The name of the check")
    branch_diff: ProposedChangeBranchDiff = Field(..., description="The calculated diff between the two branches")
    timeout: int = Field(..., description="The timeout for the check")


class TriggerRepositoryUserChecks(BaseModel):
    """Sent to trigger the user defined checks on a repository."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    source_branch: str = Field(..., description="The source branch")
    source_branch_sync_with_git: bool = Field(..., description="Indicates if the source branch should sync with git")
    target_branch: str = Field(..., description="The target branch")
    branch_diff: ProposedChangeBranchDiff = Field(..., description="The calculated diff between the two branches")


class TriggerRepositoryInternalChecks(BaseModel):
    """Sent to trigger the checks for a repository to be executed."""

    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    repository: str = Field(..., description="The unique ID of the Repository")
    source_branch: str = Field(..., description="The source branch")
    target_branch: str = Field(..., description="The target branch")


class CheckRepositoryMergeConflicts(BaseModel):
    """Runs a check to validate if there are merge conflicts for a proposed change between two branches."""

    validator_id: str = Field(..., description="The id of the validator associated with this check")
    validator_execution_id: str = Field(..., description="The id of current execution of the associated validator")
    check_execution_id: str = Field(..., description="The unique ID for the current execution of this check")
    proposed_change: str = Field(..., description="The unique ID of the Proposed Change")
    repository_id: str = Field(..., description="The unique ID of the Repository")
    repository_name: str = Field(..., description="The name of the Repository")
    source_branch: str = Field(..., description="The source branch")
    target_branch: str = Field(..., description="The target branch")


class RepositoryBranchInfo(BaseModel):
    internal_status: str


class RepositoryData(BaseModel):
    repository_id: str = Field(..., description="Id of the repository")
    repository_name: str = Field(..., description="Name of the repository")
    branches: dict[str, str] = Field(
        ...,
        description="Dictionary with the name of the branch as the key and the active commit id as the value",
    )

    branch_info: dict[str, RepositoryBranchInfo] = Field(default_factory=dict)
